#include <iostream>
using namespace std;
#include <fstream>
#include <sstream>
#include "array.h"
#include "arraysaccessor.h"
#include "timer.h"
#include "hashint.h"
#include "hashstring.h"
#include "util.h"
#include "relclustids.h"
#include "hashclause.h"
#include "hashgenclause.h"

const int TWO_PREDS_FREE_VAR = 2; //max free vars in a clause of len 2 or less
//instantiate a rule with constants if cluster supporting rule has no more than
//this number of constants
const int  CONSTANTS_CLUST_SIZE = 2;
const bool WRITE_RULES_PER_TIME_STEP = true;
const int  TIME_STEP_SEC = 24*3600;

typedef hash_map<string, Array<RelClustIdsHashArray*>*, HashString, EqualString> StringToRelClustIdsHashArrayMap;
typedef hash_map<string, Array<Array<string>*>*, HashString, EqualString>
  StringToArrayArrayStringMap;

bool storeLargerLinkToIdsOnly_ = true;
Timer timer_;
double globalStartSec_;
double globalLimitSec_;

double createClauseBeginSec_;
int    createClausePrevTimeStep_;


bool exceedTimeLimit()
{
  if (globalLimitSec_ <= 0.0) return false;
  double elapsedSec = timer_.time() - globalStartSec_;
  //cout << "ELAPSED TIME " << elapsedSec << " sec " << " , "
  //     << "LIMIT " << globalLimitSec_ << " sec " << endl;
  return (elapsedSec > globalLimitSec_);
}

void readRelStr(string str, string& relStr, Array<string>& typeStrs)
{
  string::size_type ltparen = str.find("(");
  assert(ltparen != string::npos);
  relStr = Util::trim( str.substr(0, ltparen) );
  int i = ltparen+1;
  while (true)
  {
    string::size_type comma = str.find(",",i);
    if (comma == string::npos) break;
    string typeStr = Util::trim( str.substr(i, comma-i) );
    typeStrs.append(typeStr);
    i = comma+1;
  }
  string::size_type rtparen = str.find(")",i);
  assert(rtparen != string::npos);
  string typeStr = Util::trim( str.substr(i, rtparen-i) );
  typeStrs.append(typeStr);
}

bool hasRepeatedTypes(const Array<string>& typeStrs)
{
  for (int i = 0; i < typeStrs.size(); i++)
    for (int j = i+1; j < typeStrs.size(); j++)
      if (typeStrs[i].compare(typeStrs[j]) == 0) return true;
  return false;
}

void readInfoFile(const char* const& infoFile, StringToStringArrayMap& relToTypesMap,
                  Array<string>& relNames, StringToIntMap& relNameToIdMap,
                  Array<RelClustIdsSet>& relClustIdsByRel,
                  StringToBoolMap& relNameToHasRepeatTypesMap)
{
  ifstream in(infoFile);
  Util::assertt(!in.fail(), "failed to open", infoFile, -1);
  string buf;
  bool start = false;
  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//Relations")  == 0) start = true;
    else if (buf.find("//Types") == 0) start = false;
    if (start)
    {
      if (buf.find("//") == 0) continue;
      string relStr;
      Array<string>* typeStrs = new Array<string>;
      readRelStr(buf, relStr, *typeStrs);

      cout << "Declared relation: " << relStr << "(" << (*typeStrs)[0];
      for (int i = 1; i < typeStrs->size(); i++) cout << "," << (*typeStrs)[i];
      cout << ")" << endl;

      if (relNameToIdMap.find(relStr) == relNameToIdMap.end())
      {
        relNameToIdMap[relStr] = relNames.size();
        relNames.append(relStr);
        assert(relToTypesMap.find(relStr) == relToTypesMap.end());
        relToTypesMap[relStr] = typeStrs;
        relNameToHasRepeatTypesMap[relStr] = hasRepeatedTypes(*typeStrs);
      }
      else
      {
        delete typeStrs;
        cout << "ERROR: duplicate declaration of " << relStr << endl; exit(-1);
      }
    }
  }

  in.close();

  relClustIdsByRel.growToSize(relNames.size());
}

bool checkHasTrueTuple(const Array<int>& clustIds,
                       const string& relName,
                       const StringToBoolMap& relNameToHasRepeatedTypesMap,
                       const Array<string>& trueTuples,
                       const Array<string>& types)
{
  StringToBoolMap::const_iterator itt = relNameToHasRepeatedTypesMap.find(relName);
  assert(itt != relNameToHasRepeatedTypesMap.end());
  //if no repeated types, then definitely supported by true tuple
  if ((*itt).second == false) return true;

  assert(types.size() == clustIds.size());
  IntToIntHashArrayMap clustIdToRepeatedPos;
  for (int i = 0; i < clustIds.size(); i++)
    for (int j = i+1; j < clustIds.size(); j++)
    {
      if (types[i].compare(types[j]) != 0) continue;
      if (clustIds[i] == clustIds[j])
      {
        int clustId = clustIds[i];
        IntToIntHashArrayMap::iterator it = clustIdToRepeatedPos.find(clustId);
        IntHashArray* repeatedPos = NULL;
        if (it == clustIdToRepeatedPos.end())
        {
          repeatedPos = new IntHashArray;
          clustIdToRepeatedPos[clustId] = repeatedPos;
        }
        else
          repeatedPos = (*it).second;

        repeatedPos->append(i);
        repeatedPos->append(j);
      }
    }

  //no repeated pos, clust comb must have a true tuple
  if (clustIdToRepeatedPos.empty()) return true;

  //find a tuple that has repeated arguments
  for (int t = 0; t < trueTuples.size(); t++)
  {
    string rel; Array<string> args;
    readRelStr(trueTuples[t], rel, args);

    bool match = true;
    IntToIntHashArrayMap::iterator it = clustIdToRepeatedPos.begin();
    for (; it != clustIdToRepeatedPos.end(); it++)
    {
      IntHashArray& repeatedPos = *((*it).second); assert(repeatedPos.size()>1);
      for (int i = 1; i < repeatedPos.size(); i++)
      {
        if ( args[ repeatedPos[0] ].compare( args[ repeatedPos[i] ] ) != 0)
        { match = false; break; }
      }
      if (!match) break; //go on to the next tuple
    }
    if (match)
    {
      IntToIntHashArrayMap::iterator it = clustIdToRepeatedPos.begin();
      for (; it != clustIdToRepeatedPos.end(); it++)  delete (*it).second;
      return true;
    }
  }

  IntToIntHashArrayMap::iterator it = clustIdToRepeatedPos.begin();
  for (; it != clustIdToRepeatedPos.end(); it++)  delete (*it).second;
  return false;
}

void readSemanticNetwork(int& numClustComb,
                         const char* const& semnetFile,
                         const StringToStringArrayMap& relToTypesMap,
                         const StringToIntMap& relNameToIdMap,
                         const Array<RelClustIdsSet>& relClustIdsByRel,
                         const StringToBoolMap& relNameToHasRepeatedTypesMap,
                         StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap)
{
  numClustComb = 0;

  ifstream in(semnetFile);
  Util::assertt(!in.fail(), "failed to open", semnetFile, -1);
  string buf;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (Util::startsWith(buf,"#START_RELATION"))
    {
      istringstream iss(buf);
      string relName, noop;
      iss >> noop >> relName;

      assert(relToTypesMap.find(relName) != relToTypesMap.end());
      while (!Util::startsWith(buf,"#END_RELATION"))
      {
        buf = Util::getLineAndTrim(in);
        int clustCombId = -1;
        if (Util::startsWith(buf,"#START_COMB"))
        {
          numClustComb++;
          Array<int> clustIds;
          istringstream iss(buf);
          int numClusters;
          iss >> noop >> noop >> clustCombId >> noop >> numClusters;
          for (int i = 0; i < numClusters; i++)
          {
            buf = Util::getLineAndTrim(in);
            assert(Util::startsWith(buf,"#START_CLUSTER"));
            istringstream iss(buf);
            int clustId;
            iss >> noop >> noop >> noop >> clustId;
            clustIds.append(clustId);

            while (!Util::startsWith(buf,"#END_CLUSTER")) //read all symbol
              buf = Util::getLineAndTrim(in);
          }
          buf = Util::getLineAndTrim(in);
          assert(Util::startsWith(buf,"#END_COMB"));

          //while (Util::getLineAndTrim(in).compare("#END_TRUE_TUPLES")!=0);
          buf = Util::getLineAndTrim(in);  Util::assertt(Util::startsWith(buf,"#START_TRUE_TUPLES"),"expect #START_TRUE_TUPLES", -1);

          //read the true tuples
          Array<string> trueTuples;
          buf = Util::getLineAndTrim(in);
          while (buf.find("#END_TRUE_TUPLES")!=0)
          {
            trueTuples.append(buf);
            buf = Util::getLineAndTrim(in);
          }

          StringToStringArrayMap::const_iterator it = relToTypesMap.find(relName);
          const Array<string>& types = *( (*it).second );

          //commented out: cites(x,x) still useful
          //if r(x,x) is not supported by a true tuple, ignore it
          if (!checkHasTrueTuple(clustIds, relName,relNameToHasRepeatedTypesMap,
                                 trueTuples, types)) continue;

          //cout << relName << "(" <<  clustIds[0];
          //for (int i = 1; i < clustIds.size(); i++)
          //  cout << "," << clustIds[i];
          //cout << ")" << endl;

          ////////// create a new RelClustIds, i.e., cluster combination

          assert(types.size() == clustIds.size());
          for (int i = 0; i < types.size(); i++)
          {
            string type = types[i];
            int clustId = clustIds[i];

            Array<RelClustIdsHashArray*>* clustIdToRelClustIdsMap;
            StringToRelClustIdsHashArrayMap::iterator iit = typeToClustIdToRelClustIdsMap.find(type);
            if (iit != typeToClustIdToRelClustIdsMap.end()) clustIdToRelClustIdsMap = (*iit).second;
            else
            {
              clustIdToRelClustIdsMap = new Array<RelClustIdsHashArray*>;
              typeToClustIdToRelClustIdsMap[type] = clustIdToRelClustIdsMap;
            }

            if (clustIdToRelClustIdsMap->size()-1 < clustId)
              clustIdToRelClustIdsMap->growToSize(clustId+1, NULL);
            if ((*clustIdToRelClustIdsMap)[clustId] == NULL)
              (*clustIdToRelClustIdsMap)[clustId] = new RelClustIdsHashArray;

            RelClustIdsHashArray& relClustIds = *( (*clustIdToRelClustIdsMap)[clustId] );

            StringToIntMap::const_iterator it = relNameToIdMap.find(relName); assert(it != relNameToIdMap.end());
            int relId = (*it).second;

            RelClustIds* newRelClustIds = new RelClustIds(relId, clustIds, clustCombId);
            pair<RelClustIdsSet::iterator,bool> pr = relClustIdsByRel[relId].insert(newRelClustIds);
            if (!pr.second) //if newRelClustIds already exists
            {
              delete newRelClustIds;
              newRelClustIds = *(pr.first); //get the existing relClustIds
            }
            relClustIds.append(newRelClustIds);
          }//for each cluster in cluster combination
        }//if start of comb
      }//while not end relation
    }//if start relation
  }
  in.close();
}

void populateAllRelClustIds(Array<RelClustIds*>& allRelClustIds,
                            const int& numClustCombs,
                            const Array<RelClustIdsSet>& relClustIdsByRel)
{
  int numRelClustIds = 0;
  for (int i = 0; i < relClustIdsByRel.size(); i++)
    numRelClustIds += relClustIdsByRel[i].size();
  assert(numRelClustIds <= numClustCombs);

  //allRelClustIds.growToSize( numRelClustIds, NULL );
  allRelClustIds.growToSize( numClustCombs, NULL );

  //int idx = 0;
  for (int i = 0; i < relClustIdsByRel.size(); i++)
  {
    RelClustIdsSet& relClustIdsSet = relClustIdsByRel[i];
    RelClustIdsSet::iterator it = relClustIdsSet.begin();
    for (; it != relClustIdsSet.end(); it++)
    {
      RelClustIds* relClustIds = (*it);
      //relClustIds->id_ = idx;
      //allRelClustIds[idx++] = relClustIds;
      int id = relClustIds->clustCombId_;
      relClustIds->id_ = id;
      assert(allRelClustIds[id] == NULL);
      allRelClustIds[id] = relClustIds;
    }
  }
  //assert(idx == numRelClustIds);
}

void setRelClustIdsLinkedToIds(const Array<string>& relNames,
                               StringToStringArrayMap& relToTypesMap,
                               const Array<RelClustIds*>& allRelClustIds,
                               StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap)
{
  for (int i = 0; i < allRelClustIds.size(); i++)
  {
    RelClustIds* relClustIds = allRelClustIds[i];  if (relClustIds == NULL) continue;
    string relName = relNames[ relClustIds->relId_ ];
    const Array<int>& clustIds = relClustIds->clustIds_;
    const Array<string>& types = *( relToTypesMap[relName] );
    assert(clustIds.size() == types.size());
    IntHashArray linkToIds;
    for (int j = 0; j < types.size(); j++)
    {
      string type = types[j];
      string tmpstr = type.substr(1,type.length()-1);
      //if (tmpstr.find("dummy")==0) continue; //ignore dummy types, e.g., professor(dummy,person)
      int clustId = clustIds[j];
      const Array<RelClustIdsHashArray*>& clustIdToRelClustIds = *( typeToClustIdToRelClustIdsMap[type] );
      RelClustIdsHashArray* linkToRcids = clustIdToRelClustIds[clustId];
      if (linkToRcids == NULL) continue;
      for (int k = 0; k < linkToRcids->size(); k++)
      {
        int linkToId = (*linkToRcids)[k]->id_;
        linkToIds.append( (*linkToRcids)[k]->id_ );
        assert(linkToId >= 0);
      }
    }
    relClustIds->setLinkToIds(linkToIds);
  }
}

void printRelToTypesMap(ostream& out, StringToStringArrayMap& relToTypesMap)
{
  out << "REL_TO_TYPES_MAP:" << endl;
  StringToStringArrayMap::iterator it = relToTypesMap.begin();
  for (; it != relToTypesMap.end(); it++)
  {
    string relName = (*it).first;
    Array<string>& types = (*(*it).second);
    out << relName << "(" << types[0];
    for (int i = 1; i < types.size(); i++)
      out << "," << types[i];
    out << ")" << endl;
  }
}

void printRelNames(ostream& out, const Array<string>& relNames)
{
  out << "RELATION_NAMES:" << endl;
  for (int i = 0; i < relNames.size(); i++)
    out << i << "  " << relNames[i] << endl;
}

void printRelNameToIdMap(ostream& out, const StringToIntMap& relNameToIdMap)
{
  out << "RELNAME_TO_ID_MAP:" << endl;
  StringToIntMap::const_iterator it = relNameToIdMap.begin();
  for (; it != relNameToIdMap.end(); it++)
    out << (*it).second << "  " << (*it).first << endl;
}

void printRelClustIdsByRel(ostream& out, const Array<RelClustIdsSet>& relClustIdsByRel,
                           const Array<string>& relNames)
{
  out << "RELCLUSTIDS_BY_REL" << endl;
  for (int i = 0; i < relClustIdsByRel.size(); i++)
  {
    RelClustIdsSet& relClustIdsSet = relClustIdsByRel[i];
    RelClustIdsSet::iterator it = relClustIdsSet.begin();
    for (; it != relClustIdsSet.end(); it++)
    {
      RelClustIds* relClustIds = (*it);
      out << relNames[relClustIds->relId_] << "(" << relClustIds->clustIds_[0];
      for (int i = 1; i < relClustIds->clustIds_.size(); i++)
        out << "," << relClustIds->clustIds_[i];
      out << ")" << endl;
    }
  }
}

void printTypeToClustIdToRelClustIdsMap(ostream& out,
                                        const StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                                        const Array<string>& relNames)
{
  out << "TYPE_TO_CLUSTID_TO_RELCLUSTIDS_MAP" << endl;
  StringToRelClustIdsHashArrayMap::const_iterator it = typeToClustIdToRelClustIdsMap.begin();
  for (; it != typeToClustIdToRelClustIdsMap.end(); it++)
  {
    out << "TYPE " << (*it).first << endl;
    Array<RelClustIdsHashArray*>& relClustIdsHashArrays = *( (*it).second );
    for (int i = 0; i < relClustIdsHashArrays.size(); i++)
    {
      RelClustIdsHashArray* rcidArr = relClustIdsHashArrays[i];
      if (rcidArr == NULL) continue;
      for (int j = 0; j < rcidArr->size(); j++)
      {
        RelClustIds* rcid = (*rcidArr)[j];
        out << "clustId " << i << " : " << relNames[ rcid->relId_ ] << "(" << rcid->clustIds_[0];
        for (int k =1; k < rcid->clustIds_.size(); k++)
          out << "," << rcid->clustIds_[k];
        out << ")" << endl;
      }
    }
  }
}

void printNumClustCombs(ostream& out, const Array<RelClustIdsSet>& relClustIdsByRel,
                        const Array<string>& relNames)
{
  out << "  NUM_CLUST_COMBS:" <<endl;
  int numClustCombs = 0;
  for (int i = 0; i < relClustIdsByRel.size(); i++)
  {
    out << relNames[i] << "  #clustCombs " << relClustIdsByRel[i].size() <<endl;
    numClustCombs += relClustIdsByRel[i].size();
  }
  out << "#totalClustCombs = " << numClustCombs << endl;
}

void printNumClusts(ostream& out,
                    StringToStringArrayMap& relToTypesMap,
                    const Array<string>& relNames,
                    const Array<RelClustIdsSet>& relClustIdsByRel)
{
  StringToIntHashArrayMap typeToClustIdsMap;
  for (int i = 0; i < relClustIdsByRel.size(); i++)
  {
     const Array<string>& types = *( relToTypesMap[ relNames[i] ] );
     RelClustIdsSet::iterator it = relClustIdsByRel[i].begin();
     for (; it != relClustIdsByRel[i].end(); it++)
     {
       RelClustIds* rcid = (*it);
       const Array<int>& clustIds = rcid->clustIds_; assert(clustIds.size() == types.size());
       for (int j = 0; j < clustIds.size(); j++)
       {
         int clustId = clustIds[j];
         string type = types[j];
         IntHashArray* clustIdsArr;
         StringToIntHashArrayMap::iterator it = typeToClustIdsMap.find(type);
         if (it == typeToClustIdsMap.end())
         {
           clustIdsArr = new IntHashArray;
           clustIdsArr->append(clustId);
           typeToClustIdsMap[type] = clustIdsArr;
         }
         else
         {
           (*it).second->append(clustId);
         }
       }
     }
  }
  int total = 0;
  out << "  PRINT_NUM_CLUSTS:" << endl;
  StringToIntHashArrayMap::iterator it = typeToClustIdsMap.begin();
  for (; it != typeToClustIdsMap.end(); it++)
  {
    out << (*it).first << "  " << (*it).second->size() << endl;
    total += (*it).second->size();
    delete (*it).second;
  }
  out << "#totalClusts " << total << endl;
}

void printClauses(ostream& out, const Array<Array<int*> >& clausesByLen,
                  StringToStringArrayMap relToTypesMap,
                  const Array<string>& relNames,
                  const Array<RelClustIds*>& allRelClustIds)
{
  for (int i = 1; i < clausesByLen.size(); i++)
  {
    out << "PRINTING CLAUSES OF LENGTH " << i << "   #CLAUSES "
        << clausesByLen[i].size() << endl;
    int clauseLen = i;
    Array<int*>& clauses = clausesByLen[i];
    for (int j = 0; j < clauses.size(); j++)
    {
      int* clause = clauses[j];
      for (int k = 0; k < clauseLen; k++)
      {
        int atomId = clause[k];
        RelClustIds* relClustIds = allRelClustIds[atomId];

        string relName = relNames[ relClustIds->relId_ ];
        const Array<string>& types = *( relToTypesMap[relName] );
        const Array<int>& clustIds = relClustIds->clustIds_;
        assert(types.size() == clustIds.size());
        out << relName << "(" << types[0] << "_" << clustIds[0];
        for (int a = 1; a < clustIds.size(); a++)
          out << "," << types[a] << "_" << clustIds[a];
        out << ")  ";
      }
      out << endl;
    }
  }
}

void sortGenClauses(Array<GenClause*>& genClauses, const int& clauseLen,
                    const int& l, const int& r)
{
  GenClause** items = (GenClause**) genClauses.getItems();;

  if (l >= r) return;
  GenClause* tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    Array<int>** iclause  = items[i]->preds();
    Array<int>** lclause  = items[l]->preds();

    int cmp = 0;
    for (int a = 0; a < clauseLen; a++)
    {
      Array<int>& ipred = *( (iclause)[a] );
      Array<int>& lpred = *( (lclause)[a] );
      if (ipred[0] < lpred[0])      { cmp = -1; break; }
      else if (ipred[0] > lpred[0]) { cmp =  1; break; }
    }

    if (cmp == 0)
    {
      for (int a = 0; a < clauseLen; a++)
      {
        Array<int>& ipred = *( (iclause)[a] );
        Array<int>& lpred = *( (lclause)[a] );

        for (int b = 1; b < ipred.size(); b++)
        {
          if (ipred[b] < lpred[b])      { cmp = -1; break; }
          else if (ipred[b] > lpred[b]) { cmp =  1; break; }
        }
        if (cmp != 0) break;
      }
    }

    if (cmp < 0)
    {
      ++last;
      GenClause* tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortGenClauses(genClauses, clauseLen, l, last-1);
  sortGenClauses(genClauses, clauseLen, last+1, r);
}

void sortGenClauses(Array<Array<GenClause*> >& sortedGenClausesByLen,
                    const Array<GenClauseSet>& genClausesByLen)
{
  cout << "sorting clauses ..." << endl;
  sortedGenClausesByLen.clear();
  sortedGenClausesByLen.growToSize( genClausesByLen.size() );
  for (int i = 1; i < genClausesByLen.size(); i++)
  {
    int clauseLen = i;
    GenClauseSet& clauseSet = genClausesByLen[i];
    Array<GenClause*>& genClauses = sortedGenClausesByLen[ clauseLen ];
    assert(genClauses.empty());

    HashGenClause::setSize(clauseLen);
    for (GenClauseSet::iterator it = clauseSet.begin(); it != clauseSet.end(); it++)
      genClauses.append( (*it) );

    sortGenClauses(genClauses, clauseLen, 0, genClauses.size()-1);
  }
}

void createTypeToVarNameMap(StringToStringMap& typeToVarNameMap,
                            const StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap)
{
  StringSet varNames;
  StringToRelClustIdsHashArrayMap::const_iterator it;
  it = typeToClustIdToRelClustIdsMap.begin();
  for (; it != typeToClustIdToRelClustIdsMap.end(); it++)
  {
    string type = (*it).first;

    bool assignedVarName = false;
    for (unsigned int i = 0; i < type.length()-1; i++)
    {
      string varName = type.substr(0,i+1);
      pair<StringSet::iterator,bool> pr = varNames.insert(varName);
      if (pr.second) //if unique varName
      {
        typeToVarNameMap[type] = varName;
        assignedVarName = true;
        break;
      }
    }

    if (assignedVarName) continue;

    int cnt = 1;
    while (true)
    {
      string varName = "";
      for (int i = 0; i < cnt; i++)
        varName += type[0];
      pair<StringSet::iterator,bool> pr = varNames.insert(varName);
      if (pr.second) //if unique varName
      {
        typeToVarNameMap[type] = varName;
        break;
      }
      cnt++;
    }
  }
}

void printGenClause(ostream& out, GenClause* genClause, const int& clauseLen,
                    const Array<string>& relNames,
                    StringToStringArrayMap& relToTypesMap,
                    StringToStringMap& typeToVarNameMap,
                    const Array<string>& idToConstMap)
{
  out << "#CAND: ";
  Array<int>** preds = genClause->preds();
  for (int k = 0; k < clauseLen; k++)
  {
    Array<int>& pred = *( preds[k] );
    string relName = relNames[ pred[0] ];
    const Array<string>& types = *( relToTypesMap[relName] );
    assert(types.size() == pred.size()-1);

    string varName;
    if (pred[1] < 0) { varName = idToConstMap[ -pred[1] ]; assert(varName.length()> 0); }
    else             { varName = typeToVarNameMap[types[0]] + Util::intToString(pred[1]); }
    out << relName << "(" << varName;

    for (int a = 2; a < pred.size(); a++)
    {
      if (pred[a] < 0) { varName = idToConstMap[ -pred[a] ]; assert(varName.length()> 0); }
      else             { varName = typeToVarNameMap[types[a-1]] + Util::intToString(pred[a]); }
      out << "," << varName;
    }
    out << ")  ";
  }

  out << "#CLUST_COMB_IDS";

  /*hash_set<int> intset;
  const Array<int>& clustCombIds = genClause->clustCombIds();
  for (int i = 0; i < clustCombIds.size(); i++)
  {
    pair<hash_set<int>::iterator,bool> pr = intset.insert(clustCombIds[i]);
    if (pr.second) out << " " << clustCombIds[i];
  }*/

  StringSet sset;
  const Array<string>& clauseReps = genClause->clauseReps();
  for (int i = 0; i < clauseReps.size(); i++)
  {
    pair<StringSet::iterator,bool> pr = sset.insert(clauseReps[i]);
    if (pr.second) out << " " << clauseReps[i];
  }
  out << endl;
}

void printGenClauses(ostream& out, const int& minLen,
                     const Array<Array<GenClause*> >& genClausesByLen,
                     const Array<string>& relNames,
                     StringToStringArrayMap& relToTypesMap,
                     StringToStringMap& typeToVarNameMap,
                     const Array<string>& idToConstMap)
{
  for (int i = minLen; i < genClausesByLen.size(); i++)
  {
    out << "PRINTING GENERALIZED CLAUSES OF LENGTH " << i << "   #GEN_CLAUSES "
        <<  genClausesByLen[i].size() << endl;
    int clauseLen = i;
    Array<GenClause*>& genClauses = genClausesByLen[i];
    for (int j = 0; j < genClauses.size(); j++)
      printGenClause(out, genClauses[j], clauseLen, relNames, relToTypesMap,
                     typeToVarNameMap, idToConstMap);
  }
}

void cleanUp(StringToStringArrayMap& relToTypesMap,
             Array<RelClustIdsSet>& relClustIdsByRel,
             StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
             Array<Array<int*> >& clausesByLen,
             Array<GenClauseSet>& genClauseSetByLen,
             Array<ofstream*>& outStreams,
             Array<string>& outFiles)
{
  //cout << "cleaning up..." << endl;
  StringToStringArrayMap::iterator it = relToTypesMap.begin();
  for (; it != relToTypesMap.end(); it++) delete (*it).second;

  StringToRelClustIdsHashArrayMap::iterator it2 = typeToClustIdToRelClustIdsMap.begin();
  for (; it2 != typeToClustIdToRelClustIdsMap.end(); it2++)
  {
    Array<RelClustIdsHashArray*>* rcidArrs = (*it2).second;
    for (int i = 0; i < rcidArrs->size(); i++)
    {
      RelClustIdsHashArray* rcidArr = (*rcidArrs)[i];
      delete rcidArr;
    }
    delete rcidArrs;
  }

  for (int i = 0; i < relClustIdsByRel.size(); i++)
  {
    Array<RelClustIds*> rem( relClustIdsByRel[i].size());
    RelClustIdsSet::iterator it = relClustIdsByRel[i].begin();
    for (; it != relClustIdsByRel[i].end(); it++) rem.append((*it));
    rem.deleteItemsAndClear();
  }

  for (int i = 0; i < clausesByLen.size(); i++)
  {
    Array<int*>& clauses = clausesByLen[i];
    for (int j = 0; j < clauses.size(); j++)
      delete [] clauses[j];
  }

  for (int i = 0; i < genClauseSetByLen.size(); i++)
  {
    GenClauseSet& genClauseSet = genClauseSetByLen[i];
    Array<GenClause*> rem( genClauseSet.size() );
    HashGenClause::setSize(i);
    for (GenClauseSet::iterator it = genClauseSet.begin(); it != genClauseSet.end(); it++)
      rem.append(*it);
   for (int j = 0; j < rem.size(); j++)
    GenClause::deleteGenClause(rem[j], i);
  }

  assert(outStreams.size() == outFiles.size());
  for (int i = 0; i < outStreams.size(); i++)
  {
    outStreams[i]->close();
    delete outStreams[i];
    unlink(outFiles[i].c_str());
  }
}

void sortPreds(Array<int>** const& preds, const int&l, const int& r)
{
  Array<int>** items = preds;

  if (l >= r) return;
  Array<int>* tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    Array<int>* ipred  = items[i];
    Array<int>* lpred  = items[l];

    int cmp;
    if ((*ipred)[0] < (*lpred)[0])      cmp = -1; //compare relation ids
    else if ((*ipred)[0] > (*lpred)[0]) cmp =  1;
    else
    {
      //NOTE: generalize this
      int inumSame = 0;
      if (ipred->size() == 3 && (*ipred)[1] == (*ipred)[2]) inumSame = 1;

      int lnumSame = 0;
      if (lpred->size() == 3 && (*lpred)[1] == (*lpred)[2]) lnumSame = 1;

      if (inumSame > lnumSame)      cmp = -1;
      else if (inumSame < lnumSame) cmp =  1;
      else
      {
        cmp = 0;
        for (int a = 0; a < ipred->size(); a++)
        {
          if ((*ipred)[a] < (*lpred)[a]) { cmp = -1; break; }
          if ((*ipred)[a] > (*lpred)[a]) { cmp =  1; break; }
        }
      }
    }

    if (cmp < 0)
    {
      ++last;
      Array<int>* tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortPreds(preds, l, last-1);
  sortPreds(preds, last+1, r);
}

void renumVars(Array<int>** const & genClause, const int& clauseLen)
{
  IntToIntMap oldToNewMap;

  for (int i = 0; i < clauseLen; i++)
  {
    Array<int>& pred = *( genClause[i] );
    for (int j = 1; j < pred.size(); j++)
    {
      int oldId = pred[j];
      IntToIntMap::iterator it = oldToNewMap.find(oldId);
      int newId;
      if (it == oldToNewMap.end())
      {
        newId = oldToNewMap.size();
        oldToNewMap[oldId] = newId;
      }
      else
        newId = (*it).second;
      pred[j] = newId;
    }
  }
}

void canonicalizeGenClause(GenClause* const& genClause, const int& clauseLen,
                           const bool& toRenumVars)
{
  Array<int>** preds = genClause->preds();
  sortPreds(preds, 0, clauseLen-1);
  if (toRenumVars) renumVars(preds, clauseLen);
  sortPreds(preds, 0, clauseLen-1);
}

void getLinkToAtomIds(Array<int>& linkToAtomIds,
                      int* const& clause,
                      const int& prevLen,
                      const Array<RelClustIds*>& allRelClustIds)
{
  hash_set<int> uniqueIds;
  for (int i = 0; i < prevLen; i++)
    uniqueIds.insert(clause[i]);

  for (int i = 0; i < prevLen; i++)
  {
    int atomId = clause[i];
    RelClustIds* atom = allRelClustIds[atomId];
    const Array<int>& linkToIds = atom->linkToIds_;
    for (int j = 0; j < linkToIds.size(); j++)
    {
      int linkToId = linkToIds[j];

      if (storeLargerLinkToIdsOnly_ && linkToId < atomId) continue;

      if ( uniqueIds.find(linkToId) == uniqueIds.end() )
      {
        uniqueIds.insert(linkToId);
        linkToAtomIds.append(linkToId);
      }
    }
  }
}

int findInsertId(const int& newAtomId, int* const & ids,
                 const int& startIdx, const int& endIdx)
{
  int numIdxs = endIdx - startIdx + 1;
  if (numIdxs == 0) return 0;
  if (numIdxs == 1)
  {
    if (newAtomId > ids[startIdx]) return startIdx + 1;
    if (newAtomId < ids[startIdx]) return startIdx;
    assert(newAtomId != ids[startIdx]);
  }
  else if (numIdxs == 2)
  {
    if (ids[startIdx] < newAtomId && newAtomId < ids[endIdx]) return endIdx;
    if (newAtomId < ids[startIdx]) return startIdx;
    if (newAtomId > ids[endIdx])   return endIdx+1;
    assert(newAtomId != ids[startIdx] && newAtomId != ids[endIdx]);
  }
  else if (numIdxs > 2)
  {
    //int midIdx = (numIdxs%2==0) ? startIdx+numIdxs/2 -1 : startIdx+numIdxs/2 + 1;
    int midIdx = startIdx + numIdxs/2;
    assert(startIdx < midIdx && midIdx < endIdx);
    if (ids[midIdx] < newAtomId && newAtomId < ids[midIdx+1]) return midIdx+1;
    if (newAtomId < ids[midIdx])  return findInsertId(newAtomId, ids, startIdx, midIdx-1);
    else if (newAtomId > ids[midIdx+1]) return findInsertId(newAtomId, ids, midIdx+1, endIdx);
    assert(newAtomId != ids[midIdx] && newAtomId != ids[midIdx+1]);
  }
  return -1;
}

void countVar(int& numFreeVar, int& numVar, GenClause* const& genClause,
              const int& clauseLen)
{
  Array<int>** preds = genClause->preds();
  IntToIntMap idToCntMap;
  for (int i = 0; i < clauseLen; i++)
  {
    Array<int>& pred = *(preds[i]);
    for (int j = 1; j < pred.size(); j++)
    {
      int id = pred[j];
      if (idToCntMap.find(id) == idToCntMap.end())  idToCntMap[id] = 1;
      else                                          idToCntMap[id]++;
    }
  }
  numFreeVar = 0;
  numVar = 0;
  for (IntToIntMap::iterator it=idToCntMap.begin(); it!=idToCntMap.end(); it++)
  {
    if ((*it).second <= 1) numFreeVar++;
    numVar++;
  }
}

int* removeFirstPredFromClause(int* const& clause, const int& clauseLen)
{
  int* newClause = new int[clauseLen-1];
  for (int i = 1; i < clauseLen; i++)
    newClause[i-1] = clause[i];
  return newClause;
}

int* createNewClause(const int& newAtomId, int* const& clause,
                     const int& clauseLen)
{
  int* newClause = new int[clauseLen+1];
  for (int i = 0; i < clauseLen; i++)
    newClause[i] = clause[i];
  newClause[clauseLen] = newAtomId;
  return newClause;
}

int* createNewClauseSorted(const int& newAtomId, int* const& clause,
                           const int& clauseLen)
{
  int newLen = clauseLen+1;
  int newAtomIdx = findInsertId(newAtomId, clause, 0, clauseLen-1);
  int* newClause = new int[newLen];
  for (int j = 0; j < newLen; j++)
  {
    if (j < newAtomIdx)       newClause[j] = clause[j];
    else if (j == newAtomIdx) newClause[j] = newAtomId;
    else                      newClause[j] = clause[j-1];
  }
  return newClause;
}

void createNewClauses(int* const& clause,
                      const int& clauseLen,
                      const Array<int>& linkToAtomIds,
                      Array<int*>& newClauses, ClauseSet& clauseSet)
{
  for (int i = 0; i < linkToAtomIds.size(); i++)
  {
    int* newClause = createNewClauseSorted(linkToAtomIds[i], clause, clauseLen);

    HashClause::setSize(clauseLen + 1);
    if (clauseSet.find(newClause) == clauseSet.end())
    {
      clauseSet.insert(newClause);
      newClauses.append(newClause);
    }
    else
      delete [] newClause;
  }
}

bool hasDupPred(const IntToIntArrayMap& relIdToPosMap,
                Array<int>** const& curPermutation)
{
  IntToIntArrayMap::const_iterator iit = relIdToPosMap.begin();
  for (; iit != relIdToPosMap.end(); iit++)
  {
    Array<int>& pos = *( (*iit).second );
    for (int i = 0; i < pos.size(); i++)
      for (int j = i+1; j < pos.size(); j++)
      {
        Array<int>* ipred = curPermutation[ pos[i] ];
          Array<int>* jpred = curPermutation[ pos[j] ];
          if (ipred == jpred)  return true;
      }
  }
  return false;
}

bool hasRepeatedPred(Array<int>**const& preds, const int& clauseLen)
{
  for (int i = 1; i < clauseLen; i++)
  {
    Array<int>& prevPred = *( preds[i-1] );
    Array<int>& pred = *( preds[i] );
    if (prevPred[0] == pred[0]) return true;
  }
  return false;
}

//last three params are for printing of genClause
void createSameGenClauses(const GenClause* const & genClause,
                          const int& clauseLen,
                          GenClauseSet& sameGenClauses,
                          const Array<string>& relNames,
                          StringToStringArrayMap& relToTypesMap,
                          StringToStringMap& typeToVarNameMap)
{
  HashGenClause::setSize(clauseLen);

  //commented out: assume inserted before function is called
  //sameGenClauses.insert((GenClause*)genClause);

  if (!hasRepeatedPred(genClause->preds(), clauseLen)) return;

  Array<int>** preds = genClause->preds();
  Array<int> relIds;  //relIds in the order that they appear in clause
  IntToIntArrayMap relIdToPosMap; //position in clause where rel appears
  for (int i = 0; i < clauseLen; i++)
  {
    int relId = (*(preds[i]))[0];
    relIds.append(relId);
    Array<int>* pos = NULL;
    IntToIntArrayMap::iterator it = relIdToPosMap.find(relId);
    if (it==relIdToPosMap.end()) { pos=new Array<int>; relIdToPosMap[relId]=pos; }
    else                           pos=(*it).second;
    pos->append(i);
  }

  //map relation to predicates of that relation that appear in clause
  typedef hash_map<int, Array<Array<int>*>*, HashInt, EqualInt> IntToPredsArrayMap;
  IntToPredsArrayMap relToPredsMap;
  for (int i = 0; i < clauseLen; i++)
  {
    Array<int>* pred = preds[i];
    int relId = (*pred)[0];
    Array<Array<int>*>* predsArr = NULL;
    IntToPredsArrayMap::iterator it = relToPredsMap.find(relId);
    if (it == relToPredsMap.end())
    {
      predsArr = new Array< Array<int>* >;
      relToPredsMap[relId] = predsArr;
    }
    else
      predsArr = (*it).second;
    predsArr->append(pred);
  }

  ArraysAccessor<Array<int>*> acc;
  for (int i = 0; i < relIds.size(); i++)
    acc.appendArray( relToPredsMap[ relIds[i] ] );

  //for each permutation of predicates
  while(acc.hasNextCombination())
  {
    Array<int>* pred;
    Array<int>** curPermutation = new Array<int>*[clauseLen];
    int idx = 0;
    while (acc.nextItemInCombination(pred))
      curPermutation[idx++] = pred;
    assert(idx == clauseLen);

    if (hasDupPred(relIdToPosMap,curPermutation)) {delete curPermutation;continue;}

    //renumber the vars
    IntToIntMap oldToNewVar;
    IntToIntMap::iterator varIt;
    int varId = 0;
    for (int i = 0; i < clauseLen; i++)
    {
      curPermutation[i] = new Array<int>( *( curPermutation[i] ) ); //copy preds

      Array<int>& pred = *( curPermutation[i] );
      for (int j = 1; j < pred.size(); j++)
      {
        int oldVar = pred[j];
        int newVar;
        if ( (varIt = oldToNewVar.find(oldVar)) == oldToNewVar.end() )
        {
          oldToNewVar[oldVar] = varId;
          newVar = varId;
          varId++;
        }
        else newVar = (*varIt).second;
        pred[j] = newVar;
      }
    }

    GenClause* copy = new GenClause(curPermutation);
    canonicalizeGenClause(copy, clauseLen, false);
    pair<GenClauseSet::iterator, bool> prr = sameGenClauses.insert(copy);
    if (!prr.second) GenClause::deleteGenClause(copy, clauseLen);
  }

  IntToIntArrayMap::iterator iit =  relIdToPosMap.begin();
  for(; iit != relIdToPosMap.end(); iit++) delete (*iit).second;

  IntToPredsArrayMap::iterator it = relToPredsMap.begin();
  for(; it != relToPredsMap.end(); it++) delete (*it).second;
}

GenClause* generalizeClause(int* const& clause, const int& clauseLen,
                            StringToStringArrayMap& relToTypesMap,
                            const Array<string>& relNames,
                            const Array<RelClustIds*>& allRelClustIds)
{
  StringToIntMap clustStrToIdMap;
  Array<int>** preds = new Array<int>*[clauseLen];
  GenClause* genClause = new GenClause;

  string clauseRep = Util::intToString( allRelClustIds[clause[0]]->clustCombId_);
  for (int i = 1; i < clauseLen; i++)
    clauseRep += "-" + Util::intToString(allRelClustIds[clause[i]]->clustCombId_);
  genClause->appendClauseRep(clauseRep);

  for (int i = 0; i < clauseLen; i++)
  {
    int atomId = clause[i];
    RelClustIds* relClustIds = allRelClustIds[atomId];
    const Array<int>& clustIds = relClustIds->clustIds_;

    //commented out: storing clause str reps instead
    //genClause->appendClustCombId((relClustIds->clustCombId_));

    Array<int>* pred = new Array<int>(1+clustIds.size());
    pred->append(relClustIds->relId_);

    string relName = relNames[ relClustIds->relId_ ];
    const Array<string>& types = *( relToTypesMap[relName] );
    assert(types.size() == clustIds.size());
    string clustStr = "";
    for (int j = 0; j < clustIds.size(); j++)
    {
      char buf[10];
      sprintf(buf, "%d", clustIds[j]);
      clustStr = types[j] + buf;
      StringToIntMap::iterator it = clustStrToIdMap.find(clustStr);
      int id;
      if (it == clustStrToIdMap.end())
      {
        id = clustStrToIdMap.size();
        clustStrToIdMap[clustStr] = id;
      }
      else
        id = (*it).second;
      pred->append(id);
    }
    preds[i] = pred;
  }

  genClause->setPreds(preds);
  return genClause;
}//generalizeClause

void generalizeAndStoreClause(int* const& clause, const int& clauseLen,
                              const int& maxFreeVar, const int& maxVar,
                              StringToStringArrayMap& relToTypesMap,
                              const Array<string>& relNames,
                              const Array<RelClustIds*>& allRelClustIds,
                              StringToStringMap& typeToVarNameMap,
                              GenClauseSet& genClauseSet)
{
  GenClause* genClause = generalizeClause(clause, clauseLen, relToTypesMap,
                                          relNames, allRelClustIds);
  //Array<string> noop; cout << "newClause = \n    ";
  //printGenClause(cout, genClause, clauseLen, relNames, relToTypesMap,
  //              typeToVarNameMap, noop); cout << endl;

  int numFreeVar, numVar;
  countVar(numFreeVar, numVar, genClause, clauseLen);

  int maxxFreeVar = (clauseLen > 2) ? maxFreeVar : TWO_PREDS_FREE_VAR;

  if (numFreeVar > maxxFreeVar || numVar > maxVar)
    GenClause::deleteGenClause(genClause, clauseLen);
  else
  {
    canonicalizeGenClause(genClause, clauseLen, true);

    HashGenClause::setSize(clauseLen);

    GenClauseSet sameGenClauses;
    sameGenClauses.insert(genClause);
    //create clauses that are the same modulo the ordering of their variables

    //commented out: takes too long
    //createSameGenClauses(genClause, clauseLen, sameGenClauses);

    bool genClauseExists = false;
    GenClauseSet::iterator it = sameGenClauses.begin();
    GenClauseSet::iterator existingIt;
    for (; it != sameGenClauses.end(); it++)
    {
      GenClause* gclause = (*it);
      existingIt = genClauseSet.find(gclause);
      if (existingIt != genClauseSet.end()) { genClauseExists = true; break; }
    }

    Array<GenClause*> toBeDeleted( sameGenClauses.size() );

    if (genClauseExists)
    {
      //(*(pr.first))->appendClustCombIds( genClause->clustCombIds() );
      GenClause* existingClause = (*existingIt);
      existingClause->appendClauseReps( genClause->clauseReps() );

      GenClauseSet::iterator it = sameGenClauses.begin();
      for (; it != sameGenClauses.end(); it++)
        toBeDeleted.append(*it);
    }
    else
    {
      genClauseSet.insert(genClause);

      GenClauseSet::iterator it = sameGenClauses.begin();
      for (; it != sameGenClauses.end(); it++)
        if ((*it) != genClause) toBeDeleted.append(*it);
    }

    for (int i = 0; i < toBeDeleted.size(); i++)
      GenClause::deleteGenClause(toBeDeleted[i],clauseLen);
  }
}

void generalizeClauses(Array<GenClauseSet>& genClauseSetByLen,
                       const int& maxFreeVar, const int& maxVar,
                       const Array<Array<int*> >& clausesByLen,
                       StringToStringArrayMap& relToTypesMap,
                       const Array<string>& relNames,
                       const Array<RelClustIds*>& allRelClustIds,
                       StringToStringMap& typeToVarNameMap)
{
  genClauseSetByLen.clear();
  genClauseSetByLen.growToSize( clausesByLen.size() );
  for (int i = 1; i < clausesByLen.size(); i++)
  {
    StringToIntMap clustStrToIdMap;
    int clauseLen = i;
    GenClauseSet& genClauseSet = genClauseSetByLen[clauseLen];
    HashGenClause::setSize(clauseLen);
    Array<int*>& clauses = clausesByLen[i];

    cout << "  generalizing clauses of len " << clauseLen << "..." << endl;
    for (int j = 0; j < clauses.size(); j++) //for each clause
    {
      int* clause = clauses[j];
      generalizeAndStoreClause(clause, clauseLen, maxFreeVar, maxVar,
                               relToTypesMap, relNames, allRelClustIds,
                               typeToVarNameMap, genClauseSet);
    }//for each clause
  }//for clauses of each length

}//generalizeClauses()

void createClausesBFS(Array<Array<int*> >& clausesByLen, const int& maxLen,
                      const Array<RelClustIds*>& allRelClustIds)
{
  //for i = length 1 to maxLength
  // for each clause of length i-1;
  //   for each RelClustIds containing cluster in clause
  //     add it to clause
  //     store clause

  clausesByLen.clear();
  clausesByLen.growToSize(maxLen+1);

  //create clauses of len 1
  cout << "creating clauses of length 1..." << endl;
  for (int i = 0; i < allRelClustIds.size(); i++)
  {
    int* clause = new int[1];
    if (allRelClustIds[i]==NULL) continue;
    clause[0] = allRelClustIds[i]->id_;
    clausesByLen[1].append( clause );
  }
  clausesByLen[1].compress();
  cout << "created " << clausesByLen[1].size() << " clauses of length 1"<< endl;


  //create clauses of len 1 to maxLen
  for (int i = 2; i <= maxLen; i++)
  {
    cout << "creating clauses of length " << i << "..." << endl;
    int prevLen = i-1;
    ClauseSet clauseSet;
    Array<int*>& newClauses = clausesByLen[i];   assert(newClauses.empty());
    Array<int*>& clauses    = clausesByLen[ prevLen ];

    for (int j = 0; j < clauses.size(); j++) //for each clause of len i-1
    {
      int* clause = clauses[j];

      //find atoms that atom in clause is linked to
      Array<int> linkToAtomIds;
      getLinkToAtomIds(linkToAtomIds, clause, prevLen, allRelClustIds);

      createNewClauses(clause, prevLen, linkToAtomIds, newClauses, clauseSet);
    }
    cout << "created " << newClauses.size() << " clauses of length " << i << endl;
    if (newClauses.empty()) break;
  }
}

void sortIds(Array<int>& ids, const int&l, const int& r)
{
  int* items = (int*) ids.getItems();

  if (l >= r) return;
  int tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    if (items[i] < items[l])
    {
      ++last;
      int tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortIds(ids, l, last-1);
  sortIds(ids, last+1, r);
}

bool isSortedAscending(const Array<int>& ids)
{
  for (int i = 1; i < ids.size(); i++)
    if (ids[i-1] > ids[i]) return false;
  return true;
}

void createClausesDFS(int* const& clause, const int& clauseLen,
                      const Array<int>& linkToAtomIds, const int& maxLen,
                      const int& maxFreeVar, const int& maxVar,
                      StringToStringArrayMap& relToTypesMap,
                      const Array<string>& relNames,
                      const Array<RelClustIds*>& allRelClustIds,
                      StringToStringMap& typeToVarNameMap,
                      Array<GenClauseSet>& genClauseSetByLen)
{
  //assert(isSortedAscending(linkToAtomIds)); //commented out

  //for (int i = 0; i < clauseLen-1; i++) cout << "  ";
  //cout << "createClauseDFS len " << clauseLen << endl;

  for (int i = 0; i < linkToAtomIds.size(); i++)
  {
    //create new clause by appending atom to it
    int linkToAtomId = linkToAtomIds[i];
    int* newClause = createNewClauseSorted(linkToAtomId, clause, clauseLen);

    generalizeAndStoreClause(newClause, clauseLen+1, maxFreeVar, maxVar,
                             relToTypesMap,relNames, allRelClustIds,
                             typeToVarNameMap,genClauseSetByLen[ clauseLen+1 ]);

    /*//commented out: doesn't work. E.g., Prof(p) AdvBy(s,p) Student(s) can't
      //be formed because comb id is 3 - 41 - 19
    //find new linkToAtomIds that are larger than newAtomId from
    //old linkToAtomIds and newAtom's linkToAtomIds
    RelClustIds* rcid = allRelClustIds[linkToAtomId];
    Array<int>& candLinkToAtomIds = rcid->linkToIds_;

    hash_set<int> uniqueAtomIds;
    Array<int> newLinkToAtomIds( linkToAtomIds.size()-i-1 + candLinkToAtomIds.size() );

    if (clauseLen > 0)
    {
      for (int j = i+1; j < linkToAtomIds.size(); j++)
      {
        newLinkToAtomIds.append( linkToAtomIds[j] );
        uniqueAtomIds.insert(    linkToAtomIds[j] );
      }
    }

    for (int j = 0; j < candLinkToAtomIds.size(); j++)
    {
      int candId = candLinkToAtomIds[j];
      bool allLargerThan = true;
      for (int k = 0; k < clauseLen+1; k++)
        if (candId <= newClause[k]) { allLargerThan = false; break; }
      if (allLargerThan)
      {
        pair<hash_set<int>::iterator, bool> pr = uniqueAtomIds.insert( candId );
        if (pr.second) newLinkToAtomIds.append( candId );
      }
    }

    sortIds(newLinkToAtomIds, 0, newLinkToAtomIds.size()-1);
   */

    hash_set<int> uniqueAtomIds;
    for (int i = 0; i < clauseLen+1; i++)
      uniqueAtomIds.insert( newClause[i] ) ;
    assert(uniqueAtomIds.size() == (unsigned int) clauseLen+1);

    RelClustIds* rcid = allRelClustIds[linkToAtomId];
    Array<int>& candLinkToAtomIds = rcid->linkToIds_;
    Array<int> newLinkToAtomIds( candLinkToAtomIds.size() );
    for (int i = 0; i < candLinkToAtomIds.size(); i++)
    {
      if (uniqueAtomIds.find(candLinkToAtomIds[i]) == uniqueAtomIds.end())
        newLinkToAtomIds.append( candLinkToAtomIds[i] );
    }

    if (clauseLen+1 < maxLen)
      createClausesDFS(newClause, clauseLen+1, newLinkToAtomIds, maxLen,
                       maxFreeVar, maxVar, relToTypesMap, relNames,
                       allRelClustIds, typeToVarNameMap, genClauseSetByLen);
    else
      delete [] newClause;
  }
  delete [] clause;
}//createClausesDFS()

void createClausesDFS(const int& maxLen,const int& maxFreeVar,const int& maxVar,
                      StringToStringArrayMap& relToTypesMap,
                      const Array<string>& relNames,
                      const Array<RelClustIds*>& allRelClustIds,
                      StringToStringMap& typeToVarNameMap,
                      Array<GenClauseSet>& genClauseSetByLen)
{
  Array<int> linkToAtomIds( allRelClustIds.size() );
  for (int i = 0; i < allRelClustIds.size(); i++)
    linkToAtomIds.append( allRelClustIds[i]->id_ );

  assert(genClauseSetByLen.empty());
  genClauseSetByLen.growToSize(maxLen+1);
  createClausesDFS(NULL, 0, linkToAtomIds, maxLen, maxFreeVar, maxVar,
                   relToTypesMap, relNames, allRelClustIds,  typeToVarNameMap,
                   genClauseSetByLen);
}

bool samePreds(const GenClause* const& c0, const GenClause* const& c1,
               const int& clauseLen)
{
  Array<int>** preds0 = c0->preds();
  Array<int>** preds1 = c1->preds();
  for (int i = 0; i < clauseLen; i++)
  {
    Array<int>& pred0 = *( preds0[i] );
    Array<int>& pred1 = *( preds1[i] );
    if (pred0[0] != pred1[0]) return false;
  }
  return true;
}

//last three params are for printing genClause
void removeDupGenClauses(Array<Array<GenClause*> >& noDupGenClausesByLen,
                         const Array<Array<GenClause*> >& sortedGenClausesByLen,
                         const Array<string>& relNames,
                         StringToStringArrayMap& relToTypesMap,
                         StringToStringMap& typeToVarNameMap)
{
  double bsec = timer_.time();
  cout << "removing duplicate clauses ..." << endl;

  int numDupRemoved = 0;

  noDupGenClausesByLen.clear();
  noDupGenClausesByLen.growToSize(sortedGenClausesByLen.size());

  for (int i = 2; i < sortedGenClausesByLen.size(); i++)
  {
    int clauseLen = i;
    Array<GenClause*>& sortedClauses = sortedGenClausesByLen[i];

    GenClauseHashArray genClauseArr;
    HashGenClause::setSize(clauseLen);

    for (int j = 0; j < sortedClauses.size(); j++)
    {
      if (j == 0) { genClauseArr.append(sortedClauses[j]); continue; }

      bool haveSamePreds = samePreds(sortedClauses[j], sortedClauses[j-1],
                                     clauseLen);
      if (haveSamePreds) genClauseArr.append(sortedClauses[j]);

      //if don't prev and cur clauses don't have same preds or
      //this is the last clause
      if (!haveSamePreds || j == sortedClauses.size()-1)
      { //we have gathered clauses with same preds together
        if (genClauseArr.size() >= 2)
        {
          for (int k = 0; k < genClauseArr.size(); k++)
          {
            GenClause* curGenClause = genClauseArr[k];

            GenClauseSet sameGenClauses;
            sameGenClauses.insert(curGenClause);
            createSameGenClauses(curGenClause, clauseLen, sameGenClauses,
                                 relNames, relToTypesMap, typeToVarNameMap);

            GenClauseSet::iterator it = sameGenClauses.begin();
            for (; it != sameGenClauses.end(); it++)
            {
              GenClause* sameClause = (*it);

              if (sameClause == curGenClause) continue;
              int idx = genClauseArr.find(sameClause);
              if (idx >= 0)
              {
                if (idx <= k)
                {
                  cout << "ERROR: expect idx > k" << endl;
                  cout << "idx = " << idx << " , k = " << k << endl;
                  cout << "clause at idx" << endl;
                  Array<string> noop;
                  printGenClause(cout, genClauseArr[idx], clauseLen, relNames,
                                 relToTypesMap, typeToVarNameMap, noop);
                  cout << endl;
                  cout << "clause at k" << endl;
                  printGenClause(cout, genClauseArr[k], clauseLen, relNames,
                                 relToTypesMap, typeToVarNameMap, noop);
                  cout << endl;
                  exit(-1);
                }

                GenClause* remClause = genClauseArr.removeItemFastDisorder(idx);

                remClause = NULL;
                numDupRemoved++;
              }
            }

            //delete sameGenClauses
            Array<GenClause*> toBeDeleted( sameGenClauses.size() );
            for (it = sameGenClauses.begin(); it != sameGenClauses.end(); it++)
              if ((*it) != curGenClause) toBeDeleted.append(*it);
            for (int i = 0; i < toBeDeleted.size(); i++)
              GenClause::deleteGenClause(toBeDeleted[i],clauseLen);
          }
        }
        for (int k = 0; k < genClauseArr.size(); k++)
          noDupGenClausesByLen[clauseLen].append( genClauseArr[k] );

        genClauseArr.clear();
        genClauseArr.append(sortedClauses[j]);
      }//we have gathered clauses with same preds together
    }//for each sorted clause
  }//for each clause len from 2 onwards

  cout << "canonicalizing..." << endl;
  for (int i = 0; i < noDupGenClausesByLen.size(); i++)
  {
    Array<GenClause*>& genClauses = noDupGenClausesByLen[i];
    for (int j = 0; j < genClauses.size(); j++)
      canonicalizeGenClause(genClauses[j], i, true);
  }

  cout << "#dupRemoved= " << numDupRemoved << endl;
  cout << "remove duplicate clauses took ";
  timer_.printTime(cout, timer_.time()-bsec); cout << endl;
}

void removeDupClauseReps(const Array<Array<GenClause*> >& genClausesByLen)
{
  for (int i = 0; i < genClausesByLen.size(); i++)
  {
    Array<GenClause*>& genClauses = genClausesByLen[i];
    for (int j = 0; j < genClauses.size(); j++)
    {
      GenClause* genClause = genClauses[j];
      Array<string>& clauseReps = genClause->clauseReps();
      StringHashArray uniqueReps;
      for (int k = 0; k < clauseReps.size(); k++)
        uniqueReps.append(clauseReps[k]);
      clauseReps.clear();
      for (int k = 0; k < uniqueReps.size(); k++)
        clauseReps.append(uniqueReps[k]);
    }
  }
}

void createTypeToClustersMap(StringToArrayArrayStringMap& typeToClustIdToClustMap,
                             const char* const& semnetFile,
                             StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                             StringToStringArrayMap& relToTypesMap,
                             const Array<RelClustIds*>& allRelClustIds,
                             const Array<string>& relNames)
{
  StringToRelClustIdsHashArrayMap::iterator it = typeToClustIdToRelClustIdsMap.begin();
  for (; it != typeToClustIdToRelClustIdsMap.end(); it++)
  {
    string type = (*it).first;
    typeToClustIdToClustMap[type] = new Array<Array<string>*>;
  }

  ifstream in(semnetFile);  assert(!in.fail());
  string buf;
  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("#START_COMB") == 0)
    {
      int clustCombId; string noop;
      istringstream iss(buf);
      iss >> noop >> noop >> clustCombId;

      RelClustIds* rcid = allRelClustIds[clustCombId];
      if (rcid == NULL) continue;
      assert(rcid->clustCombId_ == clustCombId);

      Array<int>& clustIds = rcid->clustIds_;
      Array<string>& types = *( relToTypesMap[ relNames[rcid->relId_] ] );
      assert(clustIds.size() == types.size());

      for (int i = 0; i < clustIds.size(); i++)
      {
        string type = types[i];
        int clustId = clustIds[i];
        buf = Util::getLineAndTrim(in); assert(buf.find("#START_CLUSTER")==0);
        int cclustId;  istringstream iss(buf);  iss >> noop >> noop >> noop >> cclustId;
        assert(clustId == cclustId);
        buf = Util::getLineAndTrim(in);
        Array<string> items;
        while(buf.find("#END_CLUSTER") != 0)
        {
          items.append(buf);
          buf = Util::getLineAndTrim(in);
        }
        if (0 < items.size() && items.size() <= CONSTANTS_CLUST_SIZE)
        {
          Array<Array<string>*>& clustIdToClustMap = *( typeToClustIdToClustMap[type] );
          if (clustIdToClustMap.size() <= clustId) clustIdToClustMap.growToSize(clustId+1,NULL);
          if (clustIdToClustMap[clustId] == NULL)
            clustIdToClustMap[clustId] = new Array<string>(items);
        }
      }
    }
  }
  in.close();

  /*itt = typeToClustIdToClustMap.begin();
  for (; itt != typeToClustIdToClustMap.end(); itt++)
  {
    string type = (*itt).first;
    Array<Array<string>*>& clustIdToClustMap = *( (*itt).second );
    for (int i = 0; i < clustIdToClustMap.size(); i++)
    {
      if (clustIdToClustMap[i] == NULL) continue;
      Array<string>& clust = *( clustIdToClustMap[i] );
      cout << type << " : clustId " << i << " : ";
      for (int j = 0; j < clust.size(); j++)
        cout << clust[j] << "  ";
      cout << endl;
    }
  }
  exit(-1);*/
}

//clauseRep is a copy
void getAtomIdsFromClauseRep(Array<int>& atomIds, string clauseRep)
{
  atomIds.clear();
  //read the clust comb ids in clauseRep
  //clust comb ids == atomIds
  for (unsigned int a = 0; a < clauseRep.length(); a++)
    if (clauseRep.at(a)=='-') clauseRep.at(a) = ' ';
  clauseRep += " ::";
  istringstream iss(clauseRep);
  string buf; iss >> buf;
  while (buf.compare("::") != 0)
  {
    atomIds.append( atoi(buf.c_str()) );
    iss >> buf;
  }
}

//clustIdsInAtoms[i] < 0 if cluster at idx i has more than 2 constants
void getClustIdsInAtoms(Array<int>& clustIdsInAtoms, Array<string>& clustTypesInAtoms,
                        const Array<int>& atomIds,
                        StringToArrayArrayStringMap& typeToClustIdToClustMap,
                        StringToStringArrayMap& relToTypesMap,
                        const Array<RelClustIds*>& allRelClustIds,
                        const Array<string>& relNames)
{
  clustIdsInAtoms.clear();
  clustTypesInAtoms.clear();
  for (int i = 0; i < atomIds.size(); i++) //for each cluster comb / atom
  {
    int atomId = atomIds[i];
    RelClustIds* rcid = allRelClustIds[atomId];  assert(rcid != NULL);

    const Array<string>& types = *( relToTypesMap[relNames[rcid->relId_]] );
    const Array<int>& clustIds = rcid->clustIds_;
    assert(types.size() == clustIds.size());

    for (int j = 0; j < clustIds.size(); j++) //for each cluster in cluster comb
    {
      int clustId = clustIds[j];
      string type = types[j];
      Array<Array<string>*>& clustIdToClustMap = *( typeToClustIdToClustMap[type] );

      //if cluster have 2 or fewer syms
      clustTypesInAtoms.append(type);
      if (clustId < clustIdToClustMap.size())
      {
        if (clustIdToClustMap[clustId] != NULL) clustIdsInAtoms.append(clustId);
        else                                    clustIdsInAtoms.append(-1);
      }
      else                                      clustIdsInAtoms.append(-1);
    }
  }
}

//returns num clusters with 2 or fewer constants
int numClustsWithConst(const Array<int>& clustIdsInAtoms)
{
  int numClusts = 0;
  for (int i = 0; i < clustIdsInAtoms.size(); i++)
    if (clustIdsInAtoms[i] >= 0) ++numClusts;
  return numClusts;
}

int getConstId(int& curConstId, const string& constant,
                StringToIntMap& constToIdMap, Array<string>& idToConstMap)
{
  int constId;
  StringToIntMap::iterator it = constToIdMap.find(constant);
  if (it == constToIdMap.end())
  {
    constId = --curConstId;
    constToIdMap[constant] = constId;
    if (idToConstMap.size() <= -curConstId) idToConstMap.growToSize(2*-curConstId+1,"");
    idToConstMap[-curConstId] = constant;
  }
  else
    constId = (*it).second;
  return constId;
}

//returns true if there are clusters with 2 or fewer constants that are present
//in all supporting clauses
bool getClustIdsWithConst(Array<int>& clustIdsInAtoms,
                          Array<string>& clustTypesInAtoms,
                          GenClause* const& genClause,
                          StringToArrayArrayStringMap& typeToClustIdToClustMap,
                          StringToStringArrayMap& relToTypesMap,
                          const Array<RelClustIds*>& allRelClustIds,
                          const Array<string>& relNames)
{
  const Array<string>& clauseReps = genClause->clauseReps();

  string clauseRep = clauseReps[0];
  Array<int> atomIds;
  getAtomIdsFromClauseRep(atomIds, clauseRep);
  clustIdsInAtoms.clear(); clustTypesInAtoms.clear();
  getClustIdsInAtoms(clustIdsInAtoms, clustTypesInAtoms, atomIds,
                     typeToClustIdToClustMap, relToTypesMap, allRelClustIds,
                     relNames);
  //num clusters with 2 or fewer constants
  int numConstClusts = numClustsWithConst(clustIdsInAtoms);

  for (int i = 1; i < clauseReps.size(); i++) //for each supporting clause
  {
    string clauseRep = clauseReps[i];
    getAtomIdsFromClauseRep(atomIds, clauseRep);
    Array<int> curClustIdsInAtoms; Array<string> noop;
    getClustIdsInAtoms(curClustIdsInAtoms, noop, atomIds,
                       typeToClustIdToClustMap, relToTypesMap, allRelClustIds,
                       relNames);
    assert(curClustIdsInAtoms.size() == clustIdsInAtoms.size());

    for (int j = 0; j < curClustIdsInAtoms.size(); j++)
    {
      if (clustIdsInAtoms[j] >= 0 && (curClustIdsInAtoms[j] != clustIdsInAtoms[j]) )
      {
        clustIdsInAtoms[j] = -1;
        --numConstClusts;
      }
    }
    if (numConstClusts == 0) break;
  }

  return (numConstClusts != 0);
}

void getClustIdsInGenClause(Array<int>& clustIdsInGenClause,
                            const GenClause* genClause, const int& clauseLen)
{
  clustIdsInGenClause.clear();
  Array<int>** preds = genClause->preds();
  for (int i = 0; i < clauseLen; i++)
  {
    Array<int>& pred = *( preds[i] );
    for (int j = 1; j < pred.size(); j++)
      clustIdsInGenClause.append( pred[j] );
  }
}

void createGenClausesWithConstants(Array<string>& idToConstMap,
                                   const Array<Array<GenClause*> >& genClausesByLen,
                                   const char* const& semnetFile,
                                   StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                                   StringToStringArrayMap& relToTypesMap,
                                   const Array<RelClustIds*>& allRelClustIds,
                                   const Array<string>& relNames,
                                   StringToStringMap& typeToVarNameMap)
{
  double begSec = timer_.time();
  cout << "creating clauses with constants" << endl;
  idToConstMap.clear();
  idToConstMap.growToSize(1000,"");
  StringToIntMap constToIdMap; //map constants to their ids
  int curConstId = 0; //constants have negative ids

  StringToArrayArrayStringMap typeToClustIdToClustMap;
  createTypeToClustersMap(typeToClustIdToClustMap, semnetFile,
                          typeToClustIdToRelClustIdsMap,
                          relToTypesMap, allRelClustIds, relNames);

  for (int i = 1; i < genClausesByLen.size(); i++) //for genClauses of each len
  {
    int clauseLen = i;
    Array<GenClause*>& genClauses = genClausesByLen[i];
    GenClauseSet newGenClauses;
    HashGenClause::setSize(clauseLen);

    for (int j = 0; j < genClauses.size(); j++) //for each genClause
    {
      GenClause* genClause = genClauses[j];
      Array<int> clustIdsInAtoms; Array<string> clustTypesInAtoms;
      bool hasConstClust=getClustIdsWithConst(clustIdsInAtoms,clustTypesInAtoms,
                                              genClause,typeToClustIdToClustMap,
                                              relToTypesMap, allRelClustIds,
                                              relNames);
      assert(clustIdsInAtoms.size() == clustTypesInAtoms.size());
      if (!hasConstClust) continue;

      Array<int> clustIdsInGenClause;
      getClustIdsInGenClause(clustIdsInGenClause, genClause, clauseLen);
      assert(clustIdsInGenClause.size() == clustIdsInAtoms.size());

      ArraysAccessor<int> acc;
      for (int k = 0; k < clustIdsInAtoms.size(); k++)
      {
        int clustId = clustIdsInAtoms[k];
        Array<int>* constantIds = new Array<int>();
        if (clustId >= 0)
        {
          string clustType = clustTypesInAtoms[k];
          Array<Array<string>*>& clustIdToClustMap = *( typeToClustIdToClustMap[clustType] );
          assert(clustId < clustIdToClustMap.size());
          Array<string>* clust = clustIdToClustMap[clustId];  assert(clust);

          for (int c = 0; c < clust->size(); c++)
          {
            string constant = (*clust)[c];
            int constId = getConstId(curConstId, constant, constToIdMap,
                                     idToConstMap);
            constantIds->append(constId);
          }
        }
        else
        {
          int clustIdInGenClause = clustIdsInGenClause[k];
          constantIds->append(clustIdInGenClause);
        }
        acc.appendArray(constantIds);
      }

      Array<int> newClustIds;
      while(acc.getNextCombination(newClustIds))
      {
        int idx = 0;
        Array<int>** preds = genClause->preds();
        Array<int>** copyPreds = new Array<int>*[clauseLen];
        for (int b = 0; b < clauseLen; b++)
        {
           copyPreds[b] = new Array<int>( *(preds[b]) );
           Array<int>& copyPred = *( copyPreds[b] );
           for (int c = 1; c < copyPred.size(); c++)
             copyPred[c] = newClustIds[idx++];
        }

        GenClause* newGenClause = new GenClause();
        newGenClause->setPreds(copyPreds);
        newGenClause->appendClauseReps(genClause->clauseReps());
        pair<GenClauseSet::iterator,bool> pr = newGenClauses.insert(newGenClause);
        if (pr.second) //if successfully inserted
        {
          //cout << "newGenClauses = " << newGenClauses.size() << endl;
          //cout << "newGenClause with constant" << endl;
          //printGenClause(cout, newGenClause, clauseLen, relNames, relToTypesMap,
          //               typeToVarNameMap, idToConstMap); cout << endl;
        }
        else
        {
          GenClause* existingGenClause = *(pr.first);

          //cout << "existingGenClause " << endl;
          //printGenClause(cout, existingGenClause, clauseLen, relNames, relToTypesMap,
          //               typeToVarNameMap, idToConstMap); cout << endl;

          Util::assertt(false, "do not expect to have dup clause with constants",-1);
          existingGenClause->appendClauseReps(genClause->clauseReps());
          GenClause::deleteGenClause(newGenClause, clauseLen);
         }
      }
      acc.deleteArraysAndClear();

    }// for each genClause
    cout << "#NEW GEN CLAUSES WITH CONSTANTS OF LEN " << clauseLen << " = "
         << newGenClauses.size() << endl;
    GenClauseSet::iterator it = newGenClauses.begin();
    for(; it != newGenClauses.end(); it++)
      genClauses.append(*it);
  }//for genClauses of each length

  //clean up
  StringToArrayArrayStringMap::iterator itt = typeToClustIdToClustMap.begin();
  for (; itt != typeToClustIdToClustMap.end(); itt++)
  {
    Array<Array<string>*>* clustIdToClustMap = (*itt).second;
    clustIdToClustMap->deleteItemsAndClear();
    delete clustIdToClustMap;
  }

  cout << "creating clauses with constants took ";
  Timer::printTime(cout, timer_.time()-begSec); cout << endl;
}


void writeRulesToFile(const string& candFile,
                      Array<GenClauseSet>& genClauseSetByLen,
                      const Array<string>& relNames,
                      StringToStringArrayMap& relToTypesMap,
                      StringToStringMap& typeToVarNameMap,
                      const char* const& semnetFile,
                StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                      const Array<RelClustIds*>& allRelClustIds,
                      const bool& createClausesWithConstants)
{
  Array<Array<GenClause*> > sortedGenClausesByLen;
  sortGenClauses(sortedGenClausesByLen, genClauseSetByLen);

  Array<Array<GenClause*> > noDupGenClausesByLen;
  removeDupGenClauses(noDupGenClausesByLen, sortedGenClausesByLen,
                      relNames, relToTypesMap, typeToVarNameMap);
  removeDupClauseReps(noDupGenClausesByLen);

  Array<string> idToConstMap;
  if (createClausesWithConstants)
    createGenClausesWithConstants(idToConstMap, noDupGenClausesByLen,semnetFile,
                                  typeToClustIdToRelClustIdsMap, relToTypesMap,
                                  allRelClustIds, relNames, typeToVarNameMap);

  ofstream candOut(candFile.c_str());
  printGenClauses(candOut, 1, noDupGenClausesByLen, relNames, relToTypesMap,
                  typeToVarNameMap, idToConstMap);
  candOut.close();
}

void createClausesDFS2(int* const& clause, const int& clauseLen,
                       const Array<int>& linkToAtomIds, const int& maxLen,
                       const int& maxFreeVar, const int& maxVar,
                       StringToStringArrayMap& relToTypesMap,
                       const Array<string>& relNames,
                       const Array<RelClustIds*>& allRelClustIds,
                       StringToStringMap& typeToVarNameMap,
                       Array<GenClauseSet>& genClauseSetByLen,
                       const string& candFile,
                       const char* const& semnetFile,
                       StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                       const bool& createClausesWithConstants)
{
  for (int i = 0; i < linkToAtomIds.size(); i++)
  {
    if (exceedTimeLimit()) { cout << "EXCEEDED TIME LIMIT" << endl; break; }

    if (WRITE_RULES_PER_TIME_STEP)
    {
      int timeStep = int( (timer_.time() - createClauseBeginSec_) / TIME_STEP_SEC );
      if (timeStep > createClausePrevTimeStep_)
      {
        createClausePrevTimeStep_ = timeStep;
        string candFile2 = candFile + "_" + Util::intToString(createClausePrevTimeStep_);
        writeRulesToFile(candFile2, genClauseSetByLen, relNames, relToTypesMap,
                         typeToVarNameMap, semnetFile,
                         typeToClustIdToRelClustIdsMap, allRelClustIds,
                         createClausesWithConstants);
      }
    }

    //create new clause by appending atom to it
    int linkToAtomId = linkToAtomIds[i];
    int* newClause = createNewClauseSorted(linkToAtomId, clause, clauseLen);

    generalizeAndStoreClause(newClause, clauseLen+1, maxFreeVar, maxVar,
                             relToTypesMap,relNames, allRelClustIds,
                             typeToVarNameMap,genClauseSetByLen[ clauseLen+1 ]);

    hash_set<int> clauseAtomIds;
    for (int i = 0; i < clauseLen+1; i++)
      clauseAtomIds.insert( newClause[i] ) ;
    assert(clauseAtomIds.size() == (unsigned int) clauseLen+1);

    RelClustIds* rcid = allRelClustIds[linkToAtomId];
    Array<int>& candLinkToAtomIds = rcid->linkToIds_;
    Array<int> newLinkToAtomIds( candLinkToAtomIds.size() );
    for (int i = 0; i < candLinkToAtomIds.size(); i++)
      if (clauseAtomIds.find(candLinkToAtomIds[i]) == clauseAtomIds.end())
        newLinkToAtomIds.append( candLinkToAtomIds[i] );

    if (clauseLen+1 < maxLen)
      createClausesDFS2(newClause, clauseLen+1, newLinkToAtomIds, maxLen,
                       maxFreeVar, maxVar, relToTypesMap, relNames,
                        allRelClustIds, typeToVarNameMap, genClauseSetByLen, candFile,
                        semnetFile, typeToClustIdToRelClustIdsMap, createClausesWithConstants);
    else
      delete [] newClause;
  }
  delete [] clause;
}//createClausesDFS2()

void createClausesDFS2(const int& maxLen,const int& maxFreeVar,
                       const int& maxVar,
                       StringToStringArrayMap& relToTypesMap,
                       const Array<string>& relNames,
                       const Array<RelClustIds*>& allRelClustIds,
                       StringToStringMap& typeToVarNameMap,
                       Array<GenClauseSet>& genClauseSetByLen,
                       const string& candFile,
                       const char* const& semnetFile,
                 StringToRelClustIdsHashArrayMap& typeToClustIdToRelClustIdsMap,
                       const bool& createClausesWithConstants)
{
  //double beginSec = timer_.time();
  //int    prevTimeStep = 0;
  createClauseBeginSec_ = timer_.time();
  createClausePrevTimeStep_ = 0;

  assert(genClauseSetByLen.empty());
  genClauseSetByLen.growToSize(maxLen+1);
  for (int i = 0; i < allRelClustIds.size(); i++)
  {
    if (allRelClustIds[i] == NULL)
    { cout << "skipping atom " << i << endl; continue; }

    if (exceedTimeLimit()) { cout << "EXCEEDED TIME LIMIT1" << endl; break; }  

    if (WRITE_RULES_PER_TIME_STEP)
    {
      int timeStep = int( (timer_.time() - createClauseBeginSec_) / TIME_STEP_SEC );
      if (timeStep > createClausePrevTimeStep_)
      {
        createClausePrevTimeStep_ = timeStep;
        string candFile2 = candFile + "_" + Util::intToString(createClausePrevTimeStep_);
        writeRulesToFile(candFile2, genClauseSetByLen, relNames, relToTypesMap,
                         typeToVarNameMap, semnetFile,
                         typeToClustIdToRelClustIdsMap, allRelClustIds,
                         createClausesWithConstants);
      }
    }

    double bsec = timer_.time();
    cout << "creating clauses starting from atom "
         << i << " / " << allRelClustIds.size() << endl;
    Array<int> linkToAtomIds;
    linkToAtomIds.append( allRelClustIds[i]->id_ );

    createClausesDFS2(NULL, 0, linkToAtomIds, maxLen, maxFreeVar, maxVar,
                      relToTypesMap, relNames, allRelClustIds, typeToVarNameMap,
                      genClauseSetByLen, candFile, semnetFile, typeToClustIdToRelClustIdsMap, createClausesWithConstants);

    cout << "  time taken   = ";
    Timer::printTime(cout, timer_.time()-bsec); cout << endl;
    cout << "  elapsed time = ";
    Timer::printTime(cout, timer_.time()-createClauseBeginSec_); cout << endl;
  }
}

void writeOutClause(int* const& clause, const int& clauseLen, ofstream& out)
{
  out << clause[0];
  for (int i = 1; i < clauseLen; i++)
    out << "  " << clause[i];
  out << endl;
}

void extendClause(const int& curAtomId,
                  const Array<ofstream*>& outStreams,
                  const Array<string>& outFiles,
                  const int& maxLen, const int& maxFreeVar, const int& maxVar,
                  StringToStringArrayMap& relToTypesMap,
                  const Array<string>& relNames,
                  const Array<RelClustIds*>& allRelClustIds,
                  StringToStringMap& typeToVarNameMap,
                  Array<GenClauseSet>& genClauseSetByLen)
{
  outStreams[curAtomId]->close(); //flush out stream

  ifstream in( outFiles[curAtomId].c_str() );
  string buf;
  while (getline(in,buf))
  {
    if (buf.empty()) continue;
    int prevClauseLen = maxLen-1;
    int* prevCreatedClause = new int[prevClauseLen];
    hash_set<int> uniqueAtomIds;
    istringstream iss(buf);
    for (int i = 0; i < prevClauseLen; i++)
    {
      iss >> prevCreatedClause[i];
      uniqueAtomIds.insert(prevCreatedClause[i]);
    }
    int lastAtomId = prevCreatedClause[prevClauseLen-1];
    Array<int>& candLinkToAtomIds = allRelClustIds[ lastAtomId ]->linkToIds_;
    Array<int> nextLinkToAtomIds( candLinkToAtomIds.size() );
    for (int i = 0; i < candLinkToAtomIds.size(); i++)
    {
      int candId = candLinkToAtomIds[i];
      if (uniqueAtomIds.find(candId) == uniqueAtomIds.end())
        nextLinkToAtomIds.append(candId);
    }

    for (int i = 0; i < nextLinkToAtomIds.size(); i++)
    {
      //create new clause by appending atom to it
      int linkToAtomId = nextLinkToAtomIds[i];
      int* newClause = createNewClause(linkToAtomId, prevCreatedClause,
                                       prevClauseLen);
      int newClauseLen = prevClauseLen+1; assert(newClauseLen == maxLen);
      generalizeAndStoreClause(newClause, newClauseLen, maxFreeVar, maxVar,
                               relToTypesMap, relNames, allRelClustIds,
                               typeToVarNameMap,genClauseSetByLen[newClauseLen]);
      delete newClause;
    }
    delete prevCreatedClause;
  }
  in.close();
  delete outStreams[curAtomId];
  outStreams[curAtomId] = new ofstream( outFiles[curAtomId].c_str(),
                                        ios_base::app );
  Util::assertt(!outStreams[curAtomId]->fail(), "failed to open",
                outFiles[curAtomId], -1);
}

void createClausesDFS3(int* const& clause, const int& clauseLen,
                       const Array<int>& linkToAtomIds,
                       const Array<ofstream*>& outStreams,
                       const Array<string>& outFiles,
                       const Array<int>& atomIdToParentMap,
                       const int& maxLen, const int& maxFreeVar,
                       const int& maxVar,
                       StringToStringArrayMap& relToTypesMap,
                       const Array<string>& relNames,
                       const Array<RelClustIds*>& allRelClustIds,
                       StringToStringMap& typeToVarNameMap,
                       Array<GenClauseSet>& genClauseSetByLen)
{
  for (int i = 0; i < linkToAtomIds.size(); i++)
  {
    //create new clause by appending atom to it
    int linkToAtomId = linkToAtomIds[i];
    int* newClause = createNewClause(linkToAtomId, clause, clauseLen);
    int newClauseLen = clauseLen+1;

    generalizeAndStoreClause(newClause, newClauseLen, maxFreeVar, maxVar,
                             relToTypesMap, relNames, allRelClustIds,
                             typeToVarNameMap, genClauseSetByLen[newClauseLen]);

    //remove first pred from clause and store it
    if (newClauseLen > 1)
    {
      int* noFirstPredClause= removeFirstPredFromClause(newClause,newClauseLen);
      generalizeAndStoreClause(noFirstPredClause, clauseLen, maxFreeVar, maxVar,
                               relToTypesMap, relNames, allRelClustIds,
                               typeToVarNameMap, genClauseSetByLen[clauseLen]);
      if (clauseLen==maxLen-1)
      {
        ofstream* out = outStreams[ noFirstPredClause[0] ];
        writeOutClause(noFirstPredClause, clauseLen, *out);
      }
      delete [] noFirstPredClause;
    }

    //extend the clause
    RelClustIds* rcid = allRelClustIds[linkToAtomId];
    Array<int>& candLinkToAtomIds = rcid->linkToIds_;
    Array<int> newLinkToAtomIds( candLinkToAtomIds.size() );

    if (newClauseLen == 1) //if unit clause
    {
      assert(linkToAtomIds.size() == 1);
      int curAtomId = linkToAtomId;

      if (atomIdToParentMap[curAtomId] >= 0) //if curAtom visited via a parent
      {
        extendClause(curAtomId, outStreams, outFiles, maxLen, maxFreeVar,maxVar,
                     relToTypesMap, relNames, allRelClustIds, typeToVarNameMap,
                     genClauseSetByLen);
        newLinkToAtomIds.append( atomIdToParentMap[curAtomId] );
      }
      else
      {
        for (int j = 0; j < candLinkToAtomIds.size(); j++)
        {
          int candId = candLinkToAtomIds[j];
          if (atomIdToParentMap[candId] < 0 && candId != curAtomId)
          {
            atomIdToParentMap[candId] = curAtomId;
            newLinkToAtomIds.append( candId );
          }
        }
      }
    }
    else
    { //if clause has two or more preds
      hash_set<int> clauseAtomIds;
      for (int i = 0; i < clauseLen+1; i++)
        clauseAtomIds.insert( newClause[i] ) ;
      assert(clauseAtomIds.size() == (unsigned int) clauseLen+1);

      for (int i = 0; i < candLinkToAtomIds.size(); i++)
        if (clauseAtomIds.find(candLinkToAtomIds[i]) == clauseAtomIds.end())
          newLinkToAtomIds.append( candLinkToAtomIds[i] );
    }

    if (newClauseLen < maxLen)
      createClausesDFS3(newClause, newClauseLen, newLinkToAtomIds,
                        outStreams, outFiles, atomIdToParentMap, maxLen,
                        maxFreeVar, maxVar, relToTypesMap, relNames,
                        allRelClustIds, typeToVarNameMap, genClauseSetByLen);
    else
      delete [] newClause;

  }
  delete [] clause;
}//createClausesDFS3()


void createClausesDFS3(const Array<ofstream*>& outStreams,
                       const Array<string>& outFiles,
                       const int& maxLen,const int& maxFreeVar,const int& maxVar,
                       StringToStringArrayMap& relToTypesMap,
                       const Array<string>& relNames,
                       const Array<RelClustIds*>& allRelClustIds,
                       StringToStringMap& typeToVarNameMap,
                       Array<GenClauseSet>& genClauseSetByLen)
{
  double beginSec = timer_.time();

  assert(genClauseSetByLen.empty());
  genClauseSetByLen.growToSize(maxLen+1);

  Array<int> atomIdToParentMap;
  atomIdToParentMap.growToSize(allRelClustIds.size(),-1);

  for (int i = 0; i < allRelClustIds.size(); i++)
  {
    if (allRelClustIds[i] == NULL)
    { cout << "3 skipping atom " << i << endl; continue; }

    cout << "3 creating clauses starting from atom "
         << i << " / " << allRelClustIds.size() << endl;
    Array<int> linkToAtomIds;
    linkToAtomIds.append( allRelClustIds[i]->id_ );

    double bsec = timer_.time();
    createClausesDFS3(NULL, 0, linkToAtomIds, outStreams, outFiles,
                      atomIdToParentMap, maxLen, maxFreeVar, maxVar,
                      relToTypesMap, relNames, allRelClustIds, typeToVarNameMap,
                      genClauseSetByLen);

    cout << "  time taken = ";
    Timer::printTime(cout, timer_.time()-bsec); cout << endl;
    cout << "  elapsed time = ";
    Timer::printTime(cout, timer_.time()-beginSec); cout << endl;
  }
}

void createOutStreams(Array<ofstream*>& outStreams,
                      Array<string>& outFiles,
                      const string& tmpDir,
                      const Array<RelClustIds*>& allRelClustIds)
{
  outStreams.clear();
  outFiles.clear();
  outStreams.growToSize( allRelClustIds.size(), NULL );
  outFiles.growToSize(   allRelClustIds.size() );

  for (int i = 0; i < allRelClustIds.size(); i++)
  {
    int atomId = allRelClustIds[i]->id_; if (allRelClustIds[i] == NULL) continue;
    Util::assertt(atomId == i, "expect atomId == i", atomId, i, -1);
    string outFile = tmpDir;
    if (outFile.at( outFile.length()-1 ) != '/') outFile += '/';
    outFile += Util::intToString(atomId) + ".clauses";

    outStreams[atomId] = new ofstream( outFile.c_str( ));
    outFiles[atomId] = outFile;
    Util::assertt(!outStreams[atomId]->fail(), "failed to open", outFile, -1);
  }
}



int main(int argc, char* argv[])
{
  if (argc != 7+1)
  {
    cout << "usage: " << argv[0]
         << " <infoFile> <semnetFile> <maxLen> <maxFreeVar> <maxVar> "
         << "<outCandFile> <timeLimitInHours>" << endl;
    exit(-1);
  }

  const char*  infoFile   = argv[1];
  const char*  semnetFile = argv[2];
  const int    maxLen     = atoi(argv[3]);
  const int    maxFreeVar = atoi(argv[4]);
  const int    maxVar     = atoi(argv[5]);
  const char*  candFile   = argv[6];
  const double hrLimit    = atof(argv[7]); //negative means no limit
  const int   dfs        = 2; //1 2 3
  const char* tmpDir     = "tmpdir/"; //for dfs3
  storeLargerLinkToIdsOnly_ = true;
  bool createClausesWithConstants = false;

  cout << "----------------------------------" << endl;
  cout << "infoFile    " << infoFile   << endl;
  cout << "semnetFile  " << semnetFile << endl;
  cout << "maxLen      " << maxLen     << endl;
  cout << "maxFreeVar  " << maxFreeVar << endl;
  cout << "maxVar      " << maxVar     << endl;
  cout << "candFile    " << candFile   << endl;
  cout << "hrLimit     " << hrLimit    << endl;
  cout << "dfs         " << dfs        << endl;
  if (dfs == 3) cout << "tmpDir      " << tmpDir     << endl;
  cout << "storeLargerLinkToIdsOnly "
       << (storeLargerLinkToIdsOnly_?"true":"false") << endl;
  cout << "----------------------------------" << endl;

  globalLimitSec_ = hrLimit * 60.0 * 60.0;
  globalStartSec_ = timer_.time();
  double startSec = timer_.time();

  StringToStringArrayMap relToTypesMap;   //maps relation name to type names
  Array<string> relNames;                 //relation names
  StringToIntMap relNameToIdMap;          //relation name to relation id
  Array<RelClustIdsSet> relClustIdsByRel; //RelClustIds contained in each rel
  Array<RelClustIds*> allRelClustIds;     //contains all RelClustIds
  StringToBoolMap relNameToHasRepeatedTypesMap;
  //type name to clustId to RelClustIdsHashArray
  StringToRelClustIdsHashArrayMap typeToClustIdToRelClustIdsMap;
  StringToStringMap typeToVarNameMap;
  Array<ofstream*> outStreams; //contains prev created clause of len maxLen-1
  Array<string> outFiles;      //contains prev created clause of len maxLen-1

  readInfoFile(infoFile, relToTypesMap, relNames, relNameToIdMap,
              relClustIdsByRel, relNameToHasRepeatedTypesMap);
  int numClustComb = 0;
  readSemanticNetwork(numClustComb, semnetFile, relToTypesMap, relNameToIdMap,
                      relClustIdsByRel, relNameToHasRepeatedTypesMap,
                      typeToClustIdToRelClustIdsMap);
  createTypeToVarNameMap(typeToVarNameMap, typeToClustIdToRelClustIdsMap);
  populateAllRelClustIds(allRelClustIds, numClustComb, relClustIdsByRel);
  setRelClustIdsLinkedToIds(relNames, relToTypesMap, allRelClustIds,
                            typeToClustIdToRelClustIdsMap);
  if (dfs == 3) createOutStreams(outStreams, outFiles, tmpDir, allRelClustIds);

  printNumClustCombs(cout, relClustIdsByRel, relNames);
  printNumClusts(cout, relToTypesMap, relNames, relClustIdsByRel);

  //printRelToTypesMap(cout, relToTypesMap);
  //printRelNames(cout, relNames);
  //printRelNameToIdMap(cout, relNameToIdMap);
  //printRelClustIdsByRel(cout, relClustIdsByRel, relNames);
  //printTypeToClustIdToRelClustIdsMap(cout, typeToClustIdToRelClustIdsMap, relNames);

  cout << "init took "; Timer::printTime(cout, timer_.time()-startSec);
  cout << endl;

  double begSec = timer_.time();

  Array<GenClauseSet> genClauseSetByLen;
  Array<Array<int*> > clausesByLen;

  if (dfs > 0)
  {
    if (dfs == 1) createClausesDFS(maxLen, maxFreeVar, maxVar, relToTypesMap,
                                   relNames, allRelClustIds, typeToVarNameMap,
                                   genClauseSetByLen);
    else
    if (dfs == 2) createClausesDFS2(maxLen, maxFreeVar, maxVar, relToTypesMap,
                                    relNames, allRelClustIds, typeToVarNameMap,
                                    genClauseSetByLen,candFile, semnetFile,
                                    typeToClustIdToRelClustIdsMap,
                                    createClausesWithConstants);
    else
    if (dfs == 3) createClausesDFS3(outStreams, outFiles, maxLen, maxFreeVar,
                                    maxVar, relToTypesMap, relNames,
                                    allRelClustIds, typeToVarNameMap,
                                    genClauseSetByLen);
    else Util::exit("unknown DFS methods", -1);
  }
  else
  {
    createClausesBFS(clausesByLen, maxLen, allRelClustIds);
    //printClauses(cout, clausesByLen, relToTypesMap, relNames, allRelClustIds);
    generalizeClauses(genClauseSetByLen, maxFreeVar, maxVar, clausesByLen,
                      relToTypesMap, relNames, allRelClustIds,typeToVarNameMap);
  }

  cout << "createClauses took "; Timer::printTime(cout, timer_.time()-begSec);
  cout << endl;

  Array<Array<GenClause*> > sortedGenClausesByLen;
  sortGenClauses(sortedGenClausesByLen, genClauseSetByLen);

  Array<Array<GenClause*> > noDupGenClausesByLen;
  removeDupGenClauses(noDupGenClausesByLen, sortedGenClausesByLen,
                      relNames, relToTypesMap, typeToVarNameMap);
  removeDupClauseReps(noDupGenClausesByLen);

  Array<string> idToConstMap;
  if (createClausesWithConstants)
    createGenClausesWithConstants(idToConstMap, noDupGenClausesByLen,semnetFile,
                                  typeToClustIdToRelClustIdsMap, relToTypesMap,
                                  allRelClustIds, relNames, typeToVarNameMap);

  printGenClauses(cout, 1, noDupGenClausesByLen, relNames, relToTypesMap,
                  typeToVarNameMap, idToConstMap);
  ofstream candOut(candFile);
  printGenClauses(candOut, 1, noDupGenClausesByLen, relNames, relToTypesMap,
                  typeToVarNameMap, idToConstMap);
  candOut.close();

  cout << "findpath took "; Timer::printTime(cout, timer_.time()-startSec);
  cout << endl;

  cleanUp(relToTypesMap, relClustIdsByRel, typeToClustIdToRelClustIdsMap,
          clausesByLen, genClauseSetByLen, outStreams, outFiles);

  return 1;
}
