#ifndef CANOPY_H_OCT_06_2008
#define CANOPY_H_OCT_06_2008

#include <climits>
#include <cassert>
#include "comb.h"

class ClusterComp
{
 public:
  static int compare(const Cluster* const& c0, const Cluster* const& c1)
  {
    if (c0 < c1) return -1;
    if (c0 > c1) return  1;
    return 0;
  }
};

typedef RBTree<Cluster*,ClusterComp> ClusterTree;
typedef ClusterTree::RBNode ClusterNode;

////////////////////////////////////////////////////////////////////////////

//this is a canopy, group all clusts A that share cluster comb B-C, i.e., A-B-C
struct Canopy
{
  Canopy() : prevNumClusters_(INT_MAX) {}

  Canopy(const Comb& comb, const int& ignoreDim) : prevNumClusters_(INT_MAX)
  {
    assert((ignoreDim >=0 && comb.size() > 1) || (ignoreDim < 0 && comb.size() >= 1));
    int size = ignoreDim >= 0 ? comb.size()-1 : comb.size();
    indexClusters_.growToSize(size);
    int cnt = 0;
    for (int i = 0; i < comb.size(); i++)
      if (i != ignoreDim) indexClusters_[cnt++] = comb[i];
    assert(cnt == indexClusters_.size());
  }

  Canopy(const Array<Cluster*>& indexClusts, const int& ignoreDim)
    : prevNumClusters_(INT_MAX)
  {
    assert((ignoreDim >=0 && indexClusts.size() > 1) || (ignoreDim < 0 && indexClusts.size() >= 1));
    int size = ignoreDim >= 0 ? indexClusts.size()-1 : indexClusts.size();
    indexClusters_.growToSize(size);
    int cnt = 0;
    for (int i = 0; i < indexClusts.size(); i++)
      if (i != ignoreDim) indexClusters_[cnt++] = indexClusts[i];
    assert(cnt == indexClusters_.size());
  }

  ~Canopy() {}

  int size() const { return clusters_.size(); }

  void add(const ClusterTree& clusters)
  {
    ClusterNode* node = clusters.head();
    while (node)
    {
      Cluster* cluster = node->data();
      clusters_.insert(cluster);
      node = node->next();
    }
  }

  bool add(Cluster* const& cluster) { return clusters_.insert(cluster); }

  bool remove(Cluster* const& cluster) { return clusters_.remove(cluster); }

  //these are the clusters that represent this canopy (serve as the index in
  //a CanopyTree
  Array<Cluster*> indexClusters_;
  ClusterTree clusters_;
  int prevNumClusters_;
};


class CanopyComp
{
 public:
  static int compare(const Canopy* const& c0, const Canopy* const& c1)
  {
    if (c0->indexClusters_.size() < c1->indexClusters_.size()) return -1;
    if (c0->indexClusters_.size() > c1->indexClusters_.size()) return  1;
    //c0 and c1 have the same number of index clusters
    const Array<Cluster*>& indexClusts0 = c0->indexClusters_;
    const Array<Cluster*>& indexClusts1 = c1->indexClusters_;
    for (int i = 0; i < indexClusts0.size(); i++)
    {
      if (indexClusts0[i] < indexClusts1[i]) return -1;
      if (indexClusts0[i] > indexClusts1[i]) return  1;
    }
    return 0;
  }
};


typedef RBTree<Canopy*,CanopyComp> CanopyTree;
typedef CanopyTree::RBNode CanopyNode;


#endif
