#ifndef CLUSTER_H_OCT_01_2008
#define CLUSTER_H_OCT_01_2008

#include <array.h>
#include "symbol.h"

class Cluster
{
 public:
  Cluster(const int& id) : id_(id), symbols_(new Array<Symbol*>) {}
  Cluster(const Cluster& clust, const int& id)
    : id_(id), symbols_(new Array<Symbol*>) { addSymbols(clust); }

  ~Cluster() { delete symbols_; }

  int size() const   { return symbols_->size(); }
  bool empty() const { return (symbols_->size()==0); }

  Array<Symbol*>& symbols() const          { return *symbols_; }
  Symbol* symbol(const int& idx) const     { return (*symbols_)[idx]; }
  Symbol* operator[](const int& idx) const { return (*symbols_)[idx]; }

  void setSymbols(Array<Symbol*>* const& s) { symbols_= s;}
  Array<Symbol*>* replaceSymbols(Array<Symbol*>* const& s)
  { Array<Symbol*>* prev = symbols_; symbols_= s; return prev; }

  int addSymbol(Symbol* const& newSymbol) { return symbols_->append(newSymbol);}
  void addSymbols(const Cluster& c)       { symbols_->append(*(c.symbols_)); }

  //add symbol to position idx and move previous symbol to end
  void addSymbol(Symbol* const& newSymbol, const int& idx)
  {
    if (idx == symbols_->size())
      addSymbol(newSymbol);
    else
    {
      Symbol* sym = (*symbols_)[idx];
      (*symbols_)[idx] = newSymbol;
      addSymbol(sym);
    }
  }

  void compress() { symbols_->compress(); }

  int id() const            { return id_; }
  void setId(const int& id) { id_ = id; }

  Symbol* removeSymbol(const int& idx)
  { return symbols_->removeItemFastDisorder(idx); }

  Symbol* removeLastSymbol() { return symbols_->removeLastItem(); }

  void clear() { symbols_->clear(); }

  ostream& print(ostream& out) const
  {
    out << "id = " << this << endl;
    out << "#symbols = " << symbols_->size() << endl;
    for (int i = 0; i < symbols_->size(); i++)
      out << i << " : " << *((*symbols_)[i]) << endl;
    return out;
  }

  void printMembers(ostream& out) const
  {
    for (int i = 0; i < symbols_->size(); i++)
      out << (*symbols_)[i]->name() << "  ";
  }

 private:
  int id_;
  Array<Symbol*>* symbols_;
};

inline
ostream& operator<<(ostream& out, const Cluster& c) { return c.print(out); }

////////////////////////////////////////////////////////////////////

class HashCluster
{
 public:
  size_t operator()(const Cluster* const& cluster) const
  {
    return hash<int>()((int)cluster);
  }
};

class EqualCluster
{
 public:
  bool operator()(const Cluster* const& c1, const Cluster* const& c2) const
  { return (c1 == c2); }
};

typedef hash_set<const Cluster*, HashCluster, EqualCluster> ConstClusterSet;

#endif
