#ifndef COMB_H_OCT_06_2008
#define COMB_H_OCT_06_2008

#include <ext/hash_set>
using namespace __gnu_cxx;
#include "cluster.h"

class Comb
{
 public:
  Comb() : numTrueTuples_(0.0), numUnknown_(0.0) {}
  Comb(const Array<Cluster*>& comb, const double& numTrueTuples)
    : comb_(comb), numTrueTuples_(numTrueTuples), numUnknown_(0.0) {}
  ~Comb() {}

  void appendCluster(Cluster* const& cluster) { comb_.append(cluster); }

  const Array<Cluster*>& comb() const { return comb_; }

  int size() const        { return comb_.size(); }
  int numClusters() const { return size(); }

  Cluster* cluster(const int& idx) const    { return comb_[idx]; }
  Cluster* operator[](const int& idx) const { return cluster(idx); }
  void setCluster(const int& idx, Cluster* const& clust) { comb_[idx] = clust; }

  void growToSize(const int& size) { comb_.growToSize(size); }
  void compress()                  { comb_.compress(); }

  double numTrueTuples() const               { return numTrueTuples_; }
  void   setNumTrueTuples(const double& t)   { numTrueTuples_ =  t;   }
  void   addNumTrueTuples(const double& t)   { numTrueTuples_ += t;   }
  void   minusNumTrueTuples(const double& t) { numTrueTuples_ -= t;   }

  double numUnknown() const             { return numUnknown_; }
  void   setNumUnknown(const double& d) { numUnknown_ =  d; }
  void   addNumUnknown(const double& d) { numUnknown_ += d; }

  double numTotalTuples() const
  {
    double total = 1.0;
    for (int d = 0; d < comb_.size(); d++)
      total *= comb_[d]->size();
    return total;
  }

  static int compare(const Comb* const& c0, const Comb* const & c1,
                     const int& ignoreDim=-1)
  { return compare(*c0, *c1, ignoreDim); }

  static int compare(const Comb& c0, const Comb& c1, const int& ignoreDim=-1)
  {
    int size = c0.size();
    if (size < c1.size()) return -1;
    if (size > c1.size()) return  1;

    for (int i = 0; i < size; i++)
    {
      if (ignoreDim >= 0 && i == ignoreDim) continue;
        //commented out: don't use id because it can change
      //if (c0[i]->id() < c1[i]->id()) { return -1; }
      //if (c0[i]->id() > c1[i]->id()) { return  1; }
      if (c0[i] < c1[i]) { return -1; }
      if (c0[i] > c1[i]) { return  1; }
      assert(c0[i] == c1[i]);
    }
    return 0;
  }

  ostream& print(ostream& out) const
  {
    for (int i = 0; i< comb_.size(); i++)
      out << comb_[i]->id() << " ";
    return out;
  }

  ostream& printFull(ostream& out) const
  {
    for (int i = 0; i< comb_.size(); i++)
    {
      Cluster* cluster = comb_[i];
      out << "clust " << i << ": ";
      for (int c = 0; c < cluster->size(); c++)
        out << (*cluster)[c]->name() << " ";
      out << endl;
    }
    return out;
  }

 private:
  Array<Cluster*> comb_; //do not delete contents; not owned by Comb
  double numTrueTuples_; //array of true tuples, one for each dim
  double numUnknown_;    //num true tuples that are set to unknown
};

inline
ostream& operator<<(ostream& out, const Comb& c) { return c.print(out); }

/////////////////////////////////////////////////////////////////////////////

class CombComp
{
 public:
  static int compare(const Comb* const& c0, const Comb* const& c1)
  {
    return Comb::compare(*c0,*c1);
  }
};


#endif
