#include "processhelper.cpp"

//extract semantic statements

//g++ -g -Wall -I./util extractstat.cpp -o extractstat


struct ClustCombNumTrue
{
  ClustCombNumTrue() : clustComb_(""), numTrueTuples_(0), totalSize_(0.0) {}
  ClustCombNumTrue(const string& ccomb, const int& ntrue, const double& tsize) 
    : clustComb_(ccomb), numTrueTuples_(ntrue), totalSize_(tsize) {}
  ~ClustCombNumTrue() {}
  string clustComb_;
  int numTrueTuples_;
  double totalSize_;
};


void createSymToClustIdMap(const Array<Array<string>*>& goldClusters,
                           StringToIntMap& symToClustIdMap)
{
  for (int i = 0; i < goldClusters.size(); i++)
  {
    Array<string>& cluster = *( goldClusters[i] );
    for (int j = 0; j < cluster.size(); j++)
    {
      string sym = cluster[j];
      assert(symToClustIdMap.find(sym) == symToClustIdMap.end());
      symToClustIdMap[sym] = i;
    }
  }
}


void makeClustCombToNumTrueTuplesMap(Array<Array<string>*>& goldClusters0,
                                     Array<Array<string>*>& goldClusters1,
                                     Array<Array<string>*>& goldClusters2,
                                     const char* const& dbFile,
                                     StringToIntMap& symToClustIdMap0,
                                     StringToIntMap& symToClustIdMap1,
                                     StringToIntMap& symToClustIdMap2,
                                    StringToIntMap& clustCombToNumTrueTuplesMap)
{
  StringToIntMap::iterator it;
  StringToIntMap::iterator rit, ait1, ait2;
  ifstream in(dbFile);
  if (in.fail()) { cout << "failed to open " << dbFile << endl; exit(-1); }
  string buf;
  while(getline(in, buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    string relStr, argStr1, argStr2;
    extractSymbols(buf, relStr, argStr1, argStr2);
    
    rit  = symToClustIdMap0.find( relStr  );
    ait1 = symToClustIdMap1.find( argStr1 );
    ait2 = symToClustIdMap2.find( argStr2 );

    int relId, argId1, argId2;
    if (rit == symToClustIdMap0.end())
    {
      Array<string>* array = new Array<string>(1);
      array->append(relStr);
      goldClusters0.append(array);
      symToClustIdMap0[relStr] = goldClusters0.size()-1;
      relId = goldClusters0.size()-1;
    }
    else relId = (*rit).second;

    if (ait1 == symToClustIdMap1.end())
    {
      Array<string>* array = new Array<string>(1);
      array->append(argStr1);
      goldClusters1.append(array);
      symToClustIdMap1[argStr1] = goldClusters1.size()-1;
      argId1 = goldClusters1.size()-1;
    }
    else argId1 = (*ait1).second;

    if (ait2 == symToClustIdMap2.end())
    {
      Array<string>* array = new Array<string>(1);
      array->append(argStr2);
      goldClusters2.append(array);
      symToClustIdMap2[argStr2] = goldClusters2.size()-1;
      argId2 = goldClusters2.size()-1;
    }
    else argId2 = (*ait2).second;

    assert(0 <= relId  && relId  <= goldClusters0.size());
    assert(0 <= argId1 && argId1 <= goldClusters1.size());
    assert(0 <= argId2 && argId2 <= goldClusters2.size());

    if (goldClusters0[relId]->size() > 1 || goldClusters1[argId1]->size() > 1 ||
        goldClusters2[argId2]->size() > 1)
    {
      char tmpbuf[1000];
      sprintf(tmpbuf, "%d %d %d", relId, argId1, argId2);
      it = clustCombToNumTrueTuplesMap.find( tmpbuf );
      if (it != clustCombToNumTrueTuplesMap.end()) (*it).second++;
      else                              clustCombToNumTrueTuplesMap[tmpbuf] = 1;
    }
  }
  in.close();  
}


void sortClustCombNumTrue(Array<ClustCombNumTrue*>& ccnt, 
                          const int& l, const int& r)
{
  if (l >= r) return;

  ClustCombNumTrue** items = (ClustCombNumTrue**) ccnt.getItems();  
  ClustCombNumTrue* tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    if (items[i]->numTrueTuples_ > items[l]->numTrueTuples_)  
    {
      ++last;
      ClustCombNumTrue* tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }
  
  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortClustCombNumTrue(ccnt, l, last-1);
  sortClustCombNumTrue(ccnt, last+1, r); 
}


void sortSampledCCombs(Array<ClustCombNumTrue*>& sampledCCombs,
                       const int& l, const int& r)
{
  if (l >= r) return;

  ClustCombNumTrue** items = (ClustCombNumTrue**) sampledCCombs.getItems();  
  ClustCombNumTrue* tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    string ccombStrI = items[i]->clustComb_;
    string ccombStrL = items[l]->clustComb_;
    istringstream issI(ccombStrI);
    istringstream issL(ccombStrL);
    int relI, arg1I, arg2I, relL, arg1L, arg2L;
    issI >> relI >> arg1I >> arg2I;  
    issL >> relL >> arg1L >> arg2L;
    
    int cmp = 0;
    if (arg1I > arg1L)      cmp =  1;
    else if (arg1I < arg1L) cmp = -1;
    else
    {      
      if (arg2I > arg2L)      cmp =  1;
      else if (arg2I < arg2L) cmp = -1;
      else
      {
        if (relI > relL)      cmp =  1;
        else if (relI < relL) cmp = -1;
        else                  cmp =  0;
      }
    }

    if (cmp < 0);
    {
      ++last;
      ClustCombNumTrue* tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }
  
  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortClustCombNumTrue(sampledCCombs, l, last-1);
  sortClustCombNumTrue(sampledCCombs, last+1, r); 
}


void printSampledCCombs(ostream& out,
                        Array<ClustCombNumTrue*>& sampledCCombs,
                        const Array<Array<string>*>& goldClusters0,
                        const Array<Array<string>*>& goldClusters1,
                        const Array<Array<string>*>& goldClusters2,
                        const char* dbFile)
{
    //for verification
  /*StringSet dbTuples;
  ifstream in(dbFile);
  if (in.fail()) { cout << "failed to open " << dbFile << endl; exit(-1); }
  string buf;
  while(getline(in, buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    dbTuples.insert(buf);
  }
  cout << "tuples read " << dbTuples.size() << endl; //*/


  sortSampledCCombs(sampledCCombs, 0, sampledCCombs.size()-1);
  for (int i = 0; i < sampledCCombs.size(); i++)
  {
    ClustCombNumTrue* ccnt = sampledCCombs[i];
    string clustCombStr = ccnt->clustComb_;
    int numTrueTuples = ccnt->numTrueTuples_;
    double totalSize  = ccnt->totalSize_;
    istringstream iss(clustCombStr);
    int relId, arg1Id, arg2Id;
    iss >> relId >> arg1Id >> arg2Id;
    
    Array<string>& clust0 = *( goldClusters0[ relId  ] );
    Array<string>& clust1 = *( goldClusters1[ arg1Id ] );
    Array<string>& clust2 = *( goldClusters2[ arg2Id ] );

    out << clustCombStr << " : numTrueTuples " << numTrueTuples << " / "
        << totalSize << endl << endl;
    int maxEle = clust0.size() > clust1.size() ? clust0.size() : clust1.size();
    maxEle = maxEle > clust2.size() ? maxEle : clust2.size();

    for (int a = 0; a < maxEle; a++)
    {
      out.width(25); out.setf(ios_base::left, ios_base::adjustfield);
      if (a < clust1.size())  out << clust1[a];  else  out << "";
      out.width(25); out.setf(ios_base::left, ios_base::adjustfield);
      if (a < clust0.size())  out << clust0[a];  else  out << "";
      out.width(25); out.setf(ios_base::left, ios_base::adjustfield);
      if (a < clust2.size())  out << clust2[a];  else  out << "";
      out << endl;
    }
    out << endl;

    //for verification
    /*int numTrue = 0;
    for (int a = 0; a < clust0.size(); a++)
      for (int b = 0; b < clust1.size(); b++)
        for (int c = 0; c < clust2.size(); c++)
        {
          string tuple = "R(";
          tuple.append( clust0[a] ).append(",").append( clust1[b] ).append(",");
          tuple.append( clust2[c] ).append(")");
          if (dbTuples.find(tuple) != dbTuples.end())
            numTrue++;
        }
    if (numTrue != numTrueTuples)
    {
      cout << "ERROR: numTrue != numTrueTuples " << numTrue << " , " 
           <<  numTrueTuples << endl;
      exit(-1);
    }//*/

  }
}


void sortClustCombNumTrue(const StringToIntMap& clustCombToNumTrueTuplesMap,
                          const Array<Array<string>*>& goldClusters0,
                          const Array<Array<string>*>& goldClusters1,
                          const Array<Array<string>*>& goldClusters2,
                          const int& minNumTrueTuples,
                          Array<ClustCombNumTrue*>& ccombs)
{
  //const int minNumTrueTuples = 2;

  StringToIntMap::const_iterator it = clustCombToNumTrueTuplesMap.begin();
  for (; it != clustCombToNumTrueTuplesMap.end(); it++)
  {
    string clustComb = (*it).first;
    int numTrueTuples = (*it).second;

    if (numTrueTuples < minNumTrueTuples) continue;

    istringstream iss(clustComb);
    int relId, arg1Id, arg2Id; 
    iss >> relId >> arg1Id >> arg2Id;
    double totalSize = goldClusters0[relId]->size()  * 
                       goldClusters1[arg1Id]->size() *
                       goldClusters2[arg2Id]->size();
    ClustCombNumTrue* ccnt = new ClustCombNumTrue(clustComb, numTrueTuples, 
                                                  totalSize);

    ccombs.append(ccnt);
  }
  cout << "number of clust combs with >=" << minNumTrueTuples 
       << " true tuples: " << ccombs.size() << endl;
  cout << "number of clust combs with < " << minNumTrueTuples 
       << " true tuples: " 
       << clustCombToNumTrueTuplesMap.size()-ccombs.size() << endl;

    //find small cluster combinations
  int numNonUnitClustComb  = 0;
  int numLargeCC = 0;  //#clustCombs with size all >2
  int numAll2 = 0; //#clustCombs with size all <=2
  int num112 = 0;
  int num113 = 0;  //#clustCombs with sizes 1,1,3 (all permutations)
  int num114 = 0;
  int num115 = 0;
  it = clustCombToNumTrueTuplesMap.begin();
  for (; it != clustCombToNumTrueTuplesMap.end(); it++)
  {
    string clustComb = (*it).first;
    //int numTrueTuples = (*it).second;

    //if (numTrueTuples < minNumTrueTuples) continue;

    istringstream iss(clustComb);
    int relId, arg1Id, arg2Id; 
    iss >> relId >> arg1Id >> arg2Id;
    int szRel  = goldClusters0[relId]->size();
    int szArg1 = goldClusters1[arg1Id]->size();
    int szArg2 = goldClusters2[arg2Id]->size();
    
    if (szRel == 1 && szArg1 == 1 && szArg2 == 1) continue;
    numNonUnitClustComb++;

    if (szRel > 2 && szArg1 > 2 && szArg2 > 2) numLargeCC++;

    if ( (szRel == 1 && szArg1 == 1 && szArg2 == 2)  ||
         (szRel == 1 && szArg1 == 2 && szArg2 == 1)  ||
         (szRel == 2 && szArg1 == 1 && szArg2 == 1) )
      num112++;

    if (szRel <= 2 && szArg1 <= 2 && szArg2 <= 2) numAll2++;
    else
    {
      int num1 = 0, num3 = 0, num4 = 0, num5 = 0;
      if (szRel == 1)   num1++;
      else if (szRel == 3) num3++;
      else if (szRel == 4) num4++;
      else if (szRel == 5) num5++;

      if (szArg1 == 1)   num1++;
      else if (szArg1 == 3) num3++;
      else if (szArg1 == 4) num4++;
      else if (szArg1 == 5) num5++;

      if (szArg2 == 1)   num1++;
      else if (szArg2 == 3) num3++;
      else if (szArg2 == 4) num4++;
      else if (szArg2 == 5) num5++;

      if (num1 == 2)
      {
        if (num3 == 1)    num113++;
        else if (num4==1) num114++;
        else if (num5==1) num115++;
      }
    }
  }
  cout << endl;
  cout<< "#NonUnitClustComb: " << numNonUnitClustComb << endl;
  cout << "#clustCombs>2>2>2: " << numLargeCC <<endl;
  cout << "#clustCombs 1-1-2: " << num112 << endl;
  cout << "#clustCombs with cluster sizes that are all <= 2: " 
       << numAll2 << endl;
  cout << "#clustCombs 1-1-3: " << num113 << endl;
  cout << "#clustCombs 1-1-4: " << num114 << endl;
  cout << "#clustCombs 1-1-5: " << num115 << endl;
  //exit(-1); //TODO: CO
  
  cout << "sorting..." << endl;
  sortClustCombNumTrue(ccombs, 0, ccombs.size()-1);
}


int main(int argc, char* argv[])
{  
  if (argc != 5+1)
  {
    cout << "usage: " << argv[0] << " <dbFile> <bestClustFile> "
         << "<minNumTrueTuples> <numSamples> <seed>" << endl;
    exit(-1);
  }

  const char* dbFile         = argv[1];
  const char* goldClustFile  = argv[2];
  const int minNumTrueTuples = atoi(argv[3]); 
  //const int minNumTrueTuples = 2; 
  const int numSamples       = atoi(argv[4]);
  const int seed             = atoi(argv[5]);
  srandom(seed);

    //read the gold clusters
  Array<Array<string>*> goldClusters0, goldClusters1, goldClusters2;
  createBestClusters(goldClustFile, 0, 0, goldClusters0);
  createBestClusters(goldClustFile, 1, 0, goldClusters1);
  createBestClusters(goldClustFile, 2, 0, goldClusters2);

    //create mapping for symbols to clusters
  StringToIntMap symToClustIdMap0, symToClustIdMap1, symToClustIdMap2;
  createSymToClustIdMap(goldClusters0, symToClustIdMap0);
  createSymToClustIdMap(goldClusters1, symToClustIdMap1);
  createSymToClustIdMap(goldClusters2, symToClustIdMap2);

    //create ClustComb to #true tuples map
  StringToIntMap clustCombToNumTrueTuplesMap;
  makeClustCombToNumTrueTuplesMap(goldClusters0, goldClusters1, goldClusters2,
                                  dbFile, symToClustIdMap0, symToClustIdMap1,
                                  symToClustIdMap2,clustCombToNumTrueTuplesMap);

  cout <<"#clustComb with >=1 true tuple: "<< clustCombToNumTrueTuplesMap.size()
       << endl;
  //exit(-1); //CO:

    //create array of ClustCombNumTrue and sort it
  Array<ClustCombNumTrue*> ccombs( clustCombToNumTrueTuplesMap.size() );
  sortClustCombNumTrue(clustCombToNumTrueTuplesMap, 
                       goldClusters0, goldClusters1, goldClusters2,
                       minNumTrueTuples, ccombs);

  //CO
  /*for (int i = 0; i < ccombs.size(); i++)
    cout  << i << " numTrue: " << ccombs[i]->numTrueTuples_ << " / " 
          << ccombs[i]->totalSize_ << endl;
    exit(-1);//*/
  
    //sample clust comb
  Array<ClustCombNumTrue*> sampledCCombs( numSamples );
  int drawn = 0;
  while (!ccombs.empty() && drawn < numSamples)
  {
    drawn++;
    ClustCombNumTrue* ccnt 
      = ccombs.removeItemFastDisorder( random() % ccombs.size() );
    sampledCCombs.append(ccnt);
  }

  printSampledCCombs(cout, sampledCCombs, goldClusters0, goldClusters1,
                     goldClusters2, dbFile);

    //clean up
  ccombs.deleteItemsAndClear(); 
  goldClusters0.deleteItemsAndClear();
  goldClusters1.deleteItemsAndClear();
  goldClusters2.deleteItemsAndClear();

  return 1;
}


