#include <iostream>
using namespace std;
#include <fstream>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <unistd.h>
#include "array.h"
#include "util.h"
#include "hashstring.h"



const int   NUM_TYPES   = 3;
const char* TYPE_NAME_0 = "relation";
const char* TYPE_NAME_1 = "concept0";
const char* TYPE_NAME_2 = "concept1";
const int   TYPE_ID_0   = 0;
const int   TYPE_ID_1   = 1;
const int   TYPE_ID_2   = 2;


class HashStringArray
{
 public:
  size_t operator()(Array<string>* const& sa) const 
  { return hash<int>()( (int) sa->size() ); }
};

class EqualStringArray
{
 public:
  bool operator()(Array<string>* const& sa1, Array<string>* const& sa2) const
  {
    if (sa1 == sa2) return true;
    if (sa1->size() != sa2->size()) return false;
    for (int i = 0; i < sa1->size(); i++)
    {
      if ( (*sa1)[i].compare( (*sa2)[i] ) != 0) return false;
    }
    return true;
  }
};

typedef hash_set<Array<string>*, HashStringArray, EqualStringArray> 
  StringArraySet;

string removeConsecutiveWhitespace(const string& buf)
{
  string sym;
  char character[2]; character[1]='\0';
  for (unsigned int i = 0; i < buf.length(); i++)
  {      
    if (buf.at(i) != '_') 
    {
      character[0] = buf.at(i);
      sym.append( character );
    }
    
    else
    {
      //replace consecutive white space with one
      if (i+1 < buf.length() && buf.at(i+1) != '_')
        sym.append("_");
    }
  }
  //cout << "SYM = " << sym << endl;
  if (sym.length() == 0) sym = "_";
  if (sym.at(0)=='_' && sym.length() > 1) sym = sym.substr(1,sym.length()-1);
  sym = Util::trim(sym);
  return sym;
}


void createBestClustersHelper(const char* const& bestClustFile,
                              const int& targetTypeId,
                              Array<Array<string>*>& bestClusters)
{
  ifstream in(bestClustFile);
  if (in.fail()) { cout << "failed to open " << bestClustFile <<endl; exit(-1);}
  string buf;
  bool inType = false;
  bool inClus = false;
  string curType; int curTypeId=-1;
  int curClustId;

  while(getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    istringstream iss(buf);
    string tok; iss >> tok;

    if (inClus)
    {
      if (tok.compare("CLUS")==0)  
      {
        iss >> curClustId;
        if (curTypeId == targetTypeId) bestClusters.append( new Array<string> );
      }
      else if (tok.compare("TYPE")==0) 
      {
        inType = false; //end of type
        inClus = false;
        int numUnitClust, numNonUnitClust;
        string typeName, noop;
        iss >> typeName >> noop >> noop >> numNonUnitClust >> noop >> noop 
            >> noop >> numUnitClust;
        assert(typeName.compare(curType)==0);
        //commented out: unit clusters may not be printed
        //assert(numUnitClust+numNonUnitClust == curClustId+1);
      }
      else  // is cluster entry
      {
        int typeId = atoi(tok.c_str());
        string cnt; string colon, sym;
        iss >> cnt >> colon >> sym;  
        if (curTypeId == targetTypeId) bestClusters.lastItem()->append(sym);
        //cout << "curType " << curType << " " << curTypeId << " , "
        //     << "sym " << sym << "  clustId " << curClustId << endl;
        assert(colon.at(0) == ':');  
        assert(typeId == curTypeId);
      }
    }
    else
    if (inType)
    {
      if (tok.compare("TYPE")==0) 
      {
        inType = false; //end of type
        inClus = false;
        int numClust;
        for (int i = 0; i < 6; i++) iss >> numClust;
        //commented out: unit clusters may not be printed
        //assert(numClust == curClustId+1);
      }
      else if (tok.compare("CLUS")==0) //start a new cluster
      {
        assert(!inClus); 
        inClus = true; 
        iss >> curClustId;
        if (curTypeId == targetTypeId) bestClusters.append( new Array<string> );
      }
      else assert(false);
    }
    else
    {  //not in a type or cluster section
      if (tok.compare("TYPE")==0) //start a new type
      { 
        inType = true; iss >> curType; 
        //cout << "reading clusters of type " << curType << endl;
        if (curType.compare(TYPE_NAME_0)==0)      curTypeId = TYPE_ID_0;
        else if (curType.compare(TYPE_NAME_1)==0) curTypeId = TYPE_ID_1;
        else if (curType.compare(TYPE_NAME_2)==0) curTypeId = TYPE_ID_2;
        else assert(false);
      }
    }

  }
  in.close();

  /*for (int j = 0; j < bestClusters.size(); j++)
  {
    cout << "\tclust " << j << endl;
    Array<string>& clust = *(bestClusters[j]);
    for (int k = 0; k < clust.size(); k++)
      cout << clust[k] << endl;
    cout << endl;
  }//*/
}//createBestClustersHelper


void createBestClusters(const char* const& bestClustFilePrefix,
                        const int& targetTypeId,
                        const int& maxLevel,
                        Array<Array<string>*>& bestClusters)
{
  StringArraySet clustersSet;
  for (int i = maxLevel; i >= 0; i--)
  {
    string bestClustFile = bestClustFilePrefix; 

    //commented out: reading single file given by bestClustFilePrefix
    //char buf[4]; sprintf(buf, "%d", i); 
    //bestClustFile.append(buf);

    cout << "reading " << bestClustFile << "..." << endl;
    Array<Array<string>*> newClusters;
    createBestClustersHelper(bestClustFile.c_str(), targetTypeId, newClusters);

    for (int i = 0; i < newClusters.size(); i++)
    {

      Array<string>* newCluster = newClusters[i];

      //sortSymbols(*newCluster, 0, newCluster->size()-1);


      if (clustersSet.find( newCluster ) == clustersSet.end())
      {
        bestClusters.append(newCluster);
        clustersSet.insert(newCluster);
      }
      else
      {
        //cout << "delete repeated clusters" << endl;        
        delete newCluster;
      }
    }
  }
}


void sort(Array<Array<string>*>& clusters, const int& l, const int& r)
{
  if (l >= r) return;

  Array<string>** items = (Array<string>**) clusters.getItems();  
  Array<string>* tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    if (items[i]->size() > items[l]->size())
    {
      ++last;
      Array<string>* tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;      
    }
  }
  
  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;

  sort(clusters, l, last-1);
  sort(clusters, last+1, r); 
}//sort

void sortSymbols(Array<string>& symbols, const int& l, const int& r)
{
  if (l >= r) return;

  string* items = (string*) symbols.getItems();  
  string tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    if (items[i].compare( items[l] ) < 0)
    {
      ++last;
      string tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;
    }
  }
  
  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;
  sortSymbols(symbols, l, last-1);
  sortSymbols(symbols, last+1, r); 
}


void extractSymbols(const string& tuple, string& relStr, string& argStr1,
                    string& argStr2)
{
  string::size_type ltparen = tuple.find(  "(" );
  string::size_type comma0  = tuple.find(  "," );
  string::size_type comma1  = tuple.rfind( "," );
  string::size_type rtparen = tuple.find(  ")" );
  relStr  = tuple.substr(ltparen+1, comma0-ltparen-1);
  argStr1 = tuple.substr(comma0 +1, comma1-comma0-1);
  argStr2 = tuple.substr(comma1 +1, rtparen-comma1-1);
  //cout << relStr << " " << argStr1 << " " << argStr2 << endl;  
}
