#ifndef SYMBOL_H_OCT_02_2008
#define SYMBOL_H_OCT_02_2008

class Cluster;

class Symbol
{
 public:
  Symbol()
    : id_(-1), name_(""),   cluster_(NULL), valid_(false) {}
  Symbol(const int& id)
    : id_(id), name_(""),   cluster_(NULL), valid_(false) {}
  Symbol(const int& id, const string& name)
    : id_(id), name_(name), cluster_(NULL), valid_(false) {}
  Symbol(const int& id, const string& name, Cluster* const& clus)
    : id_(id), name_(name), cluster_(clus), valid_(false) {}

  ~Symbol() {}

  int id() const            { return id_; }
  void setId(const int& id) { id_ = id; }

  string name() const              { return name_; }
  void setName(const string& name) { name_ = name; }

  Cluster* cluster() const                  { return cluster_; }
  void setCluster(Cluster* const & cluster) { cluster_ = cluster; }

  bool valid() const { return valid_; }
  void setValid(const bool& b)  { valid_ = b; }

  ostream& print(ostream& out) const
  {
    out << id_ << "  " << name_ << "  " << cluster_;
    return out;
  }

 private:
  int id_;
  string name_;
  Cluster* cluster_; //not owned by Symbol; do not delete
  bool valid_;
};

inline
ostream& operator<<(ostream& out, const Symbol& s)  { return s.print(out); }


class HashSymbol
{
 public:
  size_t operator()(Symbol* const& s) const { return hash<int>()(s->id()); }
};

class EqualSymbol
{
 public:
  bool operator()(Symbol* const& s1, Symbol* const& s2) const
  {
    assert( (s1->id() == s2->id() && s1 == s2) ||
            (s1->id() != s2->id() && s1 != s2) );
    return (s1==s2);
  }
};

typedef hash_map<Symbol*, int, HashSymbol, EqualSymbol> SymbolToIntMap;

#endif
