#include <iostream>
using namespace std;
#include <fstream>
#include <sstream>
#include <cstring>
#include "array.h"
#include "hashstring.h"
#include "util.h"

//g++ -g -Wall -Wno-deprecated -I/homes/gws/koks/util readandsortrules.cpp -o readandsortrules


int clauseLen(const string& clauseStr)
{
  int len = 0;
  for (unsigned int i = 0; i < clauseStr.length(); i++)
    if (clauseStr.at(i) == '(') len++;
  return len;
}


void getClausesFromMLN(Array<string>& clauses, const char* mlnFile)
{
  ifstream in(mlnFile); assert(!in.fail());
  string buf;
  bool start = false;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.compare("//function declarations") == 0) { start = true; continue;}

    if (start)
    {
      if ( buf[0] == '/' && buf[1] == '/' && buf[2] == ' ' && 
           (buf[3] == '-' || isdigit(buf[3])) )
      {
        string::size_type dblspace = buf.find("  ");
        assert( dblspace != string::npos );
        string clause = Util::trim(buf.substr(dblspace,buf.length()-dblspace));
        clauses.append(clause);
      }
    }
  }
  in.close();

  //for (int i = 0; i < clauses.size(); i++)
  //  cout << i << ": " << clauses[i] << endl;
  cout << "#clausesRead = " << clauses.size() << endl;
}


void getClausesFromMLNWithoutWts(Array<string>& clauses, const char* mlnFile)
{
  ifstream in(mlnFile); assert(!in.fail());
  string buf;
  bool start = false;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) { start = true; continue;}
    if (start) clauses.append(buf);
  }
  in.close();

  //for (int i = 0; i < clauses.size(); i++)
  //  cout << i << ": " << clauses[i] << endl;
  cout << "#clausesRead = " << clauses.size() << endl;
}

void getClauseScores(Array<double>& scores, const Array<string>& clauses,
                     const char* pruneRulesFile)
{
  StringToIntMap clauseToIdxMap;
  for (int i = 0; i < clauses.size(); i++)
    clauseToIdxMap[clauses[i]] = i;
  scores.growToSize(clauses.size(), -1);

  int cnt = 0;
  ifstream in(pruneRulesFile); assert(!in.fail());
  string buf;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("TRY_PRUNE2:") == 0)
    {
      string::size_type sp = buf.find(" ");  assert(sp != string::npos);
      string clause = Util::trim( buf.substr(sp, buf.length()-sp) );
      if (clauseToIdxMap.find(clause) != clauseToIdxMap.end())
      {
        while (buf.find("diffLogLike = ") != 0)
          getline(in,buf);
        istringstream iss(buf);
        string noop; double score;
        iss >> noop >> noop >> score;
        int idx = clauseToIdxMap[clause]; assert(idx >= 0);
        scores[idx] = score;
        cnt++;
        if (cnt == clauses.size()) break;
      }
    }    
  }
  in.close();

  for (int i = 0; i < scores.size(); i++)
  {
    //cout << i << ": " << scores[i] << endl;
    if (scores[i] < 0.0) assert(clauseLen(clauses[i]) == 1);
  }
  cout << "#scores = " << scores.size() << endl;
  assert(scores.size() == clauses.size());
}


void sortClauses(Array<string>& clauses, Array<double>& scores, 
                 const int&l, const int& r)
{
  string* items  = (string*) clauses.getItems();
  double* items2 = (double*) scores.getItems();

  if (l >= r) return;

  string tmp = items[l];
  items[l] = items[(l+r)/2];
  items[(l+r)/2] = tmp;

  double tmp2 = items2[l];
  items2[l] = items2[(l+r)/2];
  items2[(l+r)/2] = tmp2;

  int last = l;
  for (int i = l+1; i <= r; i++)
  {
    if (items2[i] > items2[l])
    {
      ++last;

      tmp = items[last];
      items[last] = items[i];
      items[i] = tmp;

      tmp2 = items2[last];
      items2[last] = items2[i];
      items2[i] = tmp2;
    }
  }

  tmp = items[l];
  items[l] = items[last];
  items[last] = tmp;

  tmp2 = items2[l];
  items2[l] = items2[last];
  items2[last] = tmp2;

  sortClauses(clauses, scores, l, last-1);
  sortClauses(clauses, scores, last+1, r);
}


void readInfoFile(Array<string>& preds, const char* infoFile)
{
  ifstream in(infoFile); assert(!in.fail());
  string buf;
  while (getline(in,buf))
  {
    buf = Util::trim(buf);
    if (buf.empty()) continue;
    if (buf.find("//") == 0) continue;
    preds.append(buf);
  }  
  in.close();
  cout << "#predDecl = " << preds.size() << endl;
}


void writeClauses(const char* outFile, const Array<string>& predDecls,
                  const Array<string>& clauses, const Array<double>& scores)
{
  ofstream out(outFile); assert(!out.fail());
  for (int i = 0; i < predDecls.size(); i++)
    out << predDecls[i] << endl;

  out << endl << endl;

  for (int i = 0; i < clauses.size(); i++)
    out << /*scores[i] << "  " <<*/ clauses[i] << endl;
  out.close();
}


int main(int argc, char* argv[])
{
  if (argc < 4+1 && argc < 5+1)
  {
    cout << "usage: " << argv[0] <<" <mln-lw file> <pruneRulesFile> <infoFile>"
         << " <outFile> [IsLwMLNFile]" << endl;
    exit(-1);
  }

  const char* mlnFile        = argv[1];
  const char* pruneRulesFile = argv[2];
  const char* infoFile       = argv[3];
  const char* outMLNFile     = argv[4];

  //MLN file is output of learnwts
  bool isLwMLNFile = true;
  if (argc == 5+1) isLwMLNFile = (strcmp(argv[5],"true")==0);

  Array<string> clauses;
  if (isLwMLNFile) getClausesFromMLN(clauses, mlnFile);
  else             getClausesFromMLNWithoutWts(clauses, mlnFile);

  Array<double> scores;
  getClauseScores(scores, clauses, pruneRulesFile);

  sortClauses(clauses, scores, 0, clauses.size()-1);

  Array<string> preds;
  readInfoFile(preds, infoFile);

  writeClauses(outMLNFile, preds, clauses, scores);
  
  return 1;

}
